# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

import logging
import multiprocessing
from enum import Enum
from functools import partial
from typing import (
    Callable,
    Dict,
    Type,
    Union,
    Optional,
    Any,
    List,
    Tuple,
    TYPE_CHECKING,
)
from uuid import uuid4

import daft
import numpy as np
import pandas as pd
import polars as pl
import pyarrow as pa
import pyarrow.dataset as ds
import pyarrow.fs
import pyarrow.parquet as papq
import ray
from ray.data.block import Block, BlockMetadata, BlockAccessor
from ray.data._internal.pandas_block import PandasBlockSchema
from ray.data.dataset import Dataset as RayDataset, MaterializedDataset
from ray.data.datasource import FilenameProvider
from ray.data.read_api import (
    from_arrow,
    from_arrow_refs,
    from_numpy,
    from_pandas,
    from_pandas_refs,
)
from tenacity import (
    Retrying,
    wait_random_exponential,
    stop_after_delay,
    retry_if_exception_type,
)

from deltacat.compute.compactor_v2.constants import MAX_RECORDS_PER_COMPACTED_FILE
from deltacat import logs
from deltacat.constants import (
    UPLOAD_SLICED_TABLE_RETRY_STOP_AFTER_DELAY,
    RETRYABLE_TRANSIENT_ERRORS,
    DOWNLOAD_MANIFEST_ENTRY_RETRY_STOP_AFTER_DELAY,
)
from deltacat.storage.model.types import (
    Dataset,
    LocalTable,
    DistributedDataset,
    LocalDataset,
)
from deltacat.storage.model.schema import SchemaConsistencyType
from deltacat.types.media import (
    DatasetType,
    DistributedDatasetType,
    ContentType,
    EXPLICIT_COMPRESSION_CONTENT_TYPES,
    ContentEncoding,
    CONTENT_TYPE_TO_EXT,
    CONTENT_ENCODING_TO_EXT,
)
from deltacat.utils import numpy as np_utils
from deltacat.utils import pandas as pd_utils
from deltacat.utils import polars as pl_utils
from deltacat.utils import pyarrow as pa_utils
from deltacat.utils.ray_utils import dataset as ds_utils
from deltacat.storage.model.manifest import (
    ManifestEntryList,
    ManifestEntry,
    ManifestMeta,
    EntryParams,
    EntryType,
    Manifest,
)
from deltacat.exceptions import (
    RetryableError,
    RetryableUploadTableError,
    NonRetryableUploadTableError,
    categorize_errors,
    RetryableDownloadTableError,
    NonRetryableDownloadTableError,
)
from deltacat.utils.common import ReadKwargsProvider
from deltacat.types.partial_download import PartialFileDownloadParams
from deltacat.utils.ray_utils.concurrency import invoke_parallel

if TYPE_CHECKING:
    from deltacat.storage.model.schema import Schema
    from deltacat.storage.model.table import Table
    from deltacat.storage.model.table_version import TableVersion

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


TABLE_TYPE_TO_READER_FUNC: Dict[str, Callable] = {
    DatasetType.PYARROW_PARQUET.value: pa_utils.file_to_parquet,
    DatasetType.PYARROW.value: pa_utils.file_to_table,
    DatasetType.PANDAS.value: pd_utils.file_to_dataframe,
    DatasetType.NUMPY.value: np_utils.file_to_ndarray,
    DatasetType.POLARS.value: pl_utils.file_to_dataframe,
}


TABLE_CLASS_TO_WRITER_FUNC: Dict[
    Type[Union[LocalTable, DistributedDataset]], Callable
] = {
    pa.Table: pa_utils.table_to_file,
    pd.DataFrame: pd_utils.dataframe_to_file,
    pl.DataFrame: pl_utils.dataframe_to_file,
    np.ndarray: np_utils.ndarray_to_file,
    RayDataset: ds_utils.dataset_to_file,
    MaterializedDataset: ds_utils.dataset_to_file,
}

TABLE_CLASS_TO_SLICER_FUNC: Dict[
    Type[Union[LocalTable, DistributedDataset]], Callable
] = {
    pa.Table: pa_utils.slice_table,
    pd.DataFrame: pd_utils.slice_dataframe,
    pl.DataFrame: pl_utils.slice_table,
    np.ndarray: np_utils.slice_ndarray,
    RayDataset: ds_utils.slice_dataset,
    MaterializedDataset: ds_utils.slice_dataset,
}

TABLE_CLASS_TO_SIZE_FUNC: Dict[
    Type[Union[LocalTable, DistributedDataset]], Callable
] = {
    pa.Table: pa_utils.table_size,
    papq.ParquetFile: pa_utils.parquet_file_size,
    pd.DataFrame: pd_utils.dataframe_size,
    pl.DataFrame: pl_utils.dataframe_size,
    np.ndarray: np_utils.ndarray_size,
    RayDataset: ds_utils.dataset_size,
    MaterializedDataset: ds_utils.dataset_size,
}

TABLE_CLASS_TO_COLUMN_NAMES_FUNC: Dict[
    Type[Union[LocalTable, DistributedDataset]], Callable
] = {
    pa.Table: lambda table: table.schema.names,
    papq.ParquetFile: lambda table: table.schema.names,
    pd.DataFrame: lambda table: table.columns.tolist(),
    pl.DataFrame: lambda table: table.columns,
    np.ndarray: lambda table: [f"{i}" for i in range(table.shape[1])],
    daft.DataFrame: lambda table: table.column_names,
    RayDataset: lambda table: table.schema().names,
    MaterializedDataset: lambda table: table.schema().names,
}

TABLE_CLASS_TO_SCHEMA_FUNC: Dict[
    Type[Union[LocalTable, DistributedDataset]], Callable
] = {
    pa.Table: lambda table: table.schema,
    papq.ParquetFile: lambda table: table.schema_arrow,
    pd.DataFrame: lambda table: pa.Schema.from_pandas(table),
    pl.DataFrame: lambda table: table.to_arrow().schema,
    np.ndarray: lambda table: pa.Schema.from_pandas(pd.DataFrame(table)),
    daft.DataFrame: lambda table: table.schema().to_pyarrow_schema(),
    RayDataset: lambda table: table.schema().base_schema,
    MaterializedDataset: lambda table: table.schema().base_schema,
}

TABLE_TYPE_TO_EMPTY_TABLE_FUNC: Dict[str, Callable] = {
    DatasetType.PYARROW.value: lambda: pa.Table.from_pydict({}),
    DatasetType.PANDAS.value: lambda: pd.DataFrame(),
    DatasetType.POLARS.value: lambda: pl.DataFrame(),
    DatasetType.NUMPY.value: lambda: np.array([]),
    DatasetType.DAFT.value: lambda: daft.DataFrame(),
    DatasetType.RAY_DATASET.value: lambda: ray.data.from_items([]),
    MaterializedDataset: lambda: ray.data.from_items([]),
}


def _numpy_array_to_pyarrow(table: np.ndarray, schema: pa.Schema) -> pa.Table:
    """Convert NumPy array to PyArrow Table via Pandas for complex type support."""
    # Convert NumPy -> Pandas -> PyArrow to handle complex types like structs
    # This follows the same path as Pandas conversion which handles all complex types properly
    pandas_df = _numpy_array_to_pandas(table, schema=schema)
    return pa.Table.from_pandas(pandas_df, schema=schema)


def _numpy_array_to_pandas(
    table: np.ndarray, *, schema: Optional[pa.Schema] = None, **kwargs
) -> pd.DataFrame:
    """Convert NumPy array to pandas DataFrame."""
    if schema and isinstance(schema, pa.Schema):
        if table.ndim == 1:
            # 1D array: single column
            column_names = [schema.names[0]] if schema.names else ["0"]
            return pd.DataFrame({column_names[0]: table}, **kwargs)
        elif table.ndim == 2:
            # 2D array: multiple columns
            column_names = (
                schema.names
                if len(schema.names) == table.shape[1]
                else [f"{i}" for i in range(table.shape[1])]
            )
            return pd.DataFrame(table, columns=column_names, **kwargs)
        else:
            raise ValueError(
                f"NumPy arrays with {table.ndim} dimensions are not supported. "
                f"Only 1D and 2D arrays are supported."
            )

    # Fallback to generic column names
    return pd.DataFrame(table, **kwargs)


def _ray_dataset_to_pyarrow(table, *, schema, **kwargs):
    """Convert Ray Dataset to PyArrow tables and concatenate."""
    arrow_refs = table.to_arrow_refs(**kwargs)
    arrow_tables = ray.get(arrow_refs)
    if len(arrow_tables) == 1:
        return arrow_tables[0]
    # Unify schemas to support schema evolution across blocks/files
    try:
        return pa.concat_tables(
            arrow_tables, promote_options="permissive", unify_schemas=True
        )
    except TypeError:
        return pa.concat_tables(arrow_tables, promote_options="permissive")


TABLE_CLASS_TO_PYARROW_FUNC: Dict[
    Type[Union[LocalTable, DistributedDataset]], Callable
] = {
    pa.Table: lambda table, *, schema, **kwargs: table,
    papq.ParquetFile: lambda table, *, schema, **kwargs: table.read(**kwargs),
    pd.DataFrame: lambda table, *, schema, **kwargs: pa.Table.from_pandas(
        table, schema=schema, preserve_index=False, **kwargs
    ),
    pl.DataFrame: lambda table, *, schema, **kwargs: pl.DataFrame.to_arrow(
        table, **kwargs
    ),
    np.ndarray: lambda table, *, schema, **kwargs: _numpy_array_to_pyarrow(
        table, schema, **kwargs
    ),
    RayDataset: _ray_dataset_to_pyarrow,
    MaterializedDataset: _ray_dataset_to_pyarrow,
    daft.DataFrame: lambda table, *, schema, **kwargs: table.to_arrow(**kwargs),
}

TABLE_CLASS_TO_PANDAS_FUNC: Dict[
    Type[Union[LocalTable, DistributedDataset]], Callable
] = {
    pa.Table: lambda table, *, schema=None, **kwargs: table.to_pandas(**kwargs),
    papq.ParquetFile: lambda table, *, schema=None, **kwargs: table.read(
        **kwargs
    ).to_pandas(**kwargs),
    pd.DataFrame: lambda table, *, schema=None, **kwargs: table,
    pl.DataFrame: lambda table, *, schema=None, **kwargs: table.to_pandas(**kwargs),
    np.ndarray: lambda table, *, schema=None, **kwargs: _numpy_array_to_pandas(
        table, schema=schema, **kwargs
    ),
    RayDataset: lambda table, *, schema=None, **kwargs: table.to_pandas(**kwargs),
    MaterializedDataset: lambda table, *, schema=None, **kwargs: table.to_pandas(
        **kwargs
    ),
    daft.DataFrame: lambda table, *, schema=None, **kwargs: table.to_pandas(**kwargs),
}


def _pyarrow_to_polars(pa_table: pa.Table, **kwargs) -> pl.DataFrame:
    """Convert PyArrow table to Polars DataFrame with clean schema."""
    # PyArrow metadata can contain invalid UTF-8 sequences that cause Polars to raise an error
    # Create a new table without metadata that might contain invalid UTF-8
    clean_schema = pa.schema(
        [
            pa.field(field.name, field.type, nullable=field.nullable)
            for field in pa_table.schema
        ]
    )
    clean_table = pa.Table.from_arrays(pa_table.columns, schema=clean_schema)
    return pl.from_arrow(clean_table, **kwargs)


def _pyarrow_to_numpy(pa_table: pa.Table, **kwargs) -> np.ndarray:
    """Convert PyArrow table to numpy array."""
    if pa_table.num_columns == 1:
        return pa_table.column(0).to_numpy(**kwargs)
    else:
        return pa_table.to_pandas().values


def _pandas_to_numpy(pd_df: pd.DataFrame, **kwargs) -> np.ndarray:
    """Convert Pandas DataFrame to numpy array."""
    if len(pd_df.columns) == 1:
        return pd_df.iloc[:, 0].to_numpy(**kwargs)
    else:
        return pd_df.values


DATASET_TYPE_FROM_PYARROW: Dict[DatasetType, Callable[[pa.Table, Dataset], Any]] = {
    DatasetType.PYARROW: lambda pa_table, **kwargs: pa_table,
    DatasetType.PANDAS: lambda pa_table, **kwargs: pa_table.to_pandas(**kwargs),
    DatasetType.POLARS: lambda pa_table, **kwargs: _pyarrow_to_polars(
        pa_table, **kwargs
    ),
    DatasetType.DAFT: lambda pa_table, **kwargs: daft.from_arrow(pa_table, **kwargs),
    DatasetType.NUMPY: lambda pa_table, **kwargs: _pyarrow_to_numpy(pa_table, **kwargs),
    DatasetType.RAY_DATASET: lambda pa_table, **kwargs: ray.data.from_arrow(pa_table),
    DatasetType.PYARROW_PARQUET: lambda pa_table, **kwargs: pa_table,  # ParquetFile is read as PyArrow Table
}


DATASET_TYPE_FROM_PANDAS: Dict[DatasetType, Callable[[pd.DataFrame, Dataset], Any]] = {
    DatasetType.PANDAS: lambda pd_df, **kwargs: pd_df,
    DatasetType.PYARROW: lambda pd_df, **kwargs: pa.Table.from_pandas(pd_df, **kwargs),
    DatasetType.POLARS: lambda pd_df, **kwargs: pl.from_pandas(pd_df, **kwargs),
    DatasetType.DAFT: lambda pd_df, **kwargs: daft.from_pandas(pd_df, **kwargs),
    DatasetType.NUMPY: lambda pd_df, **kwargs: _pandas_to_numpy(pd_df, **kwargs),
    DatasetType.RAY_DATASET: lambda pd_df, **kwargs: ray.data.from_pandas(
        pd_df, **kwargs
    ),
}


def append_column_to_parquet_file(
    parquet_file: papq.ParquetFile,
    column_name: str,
    column_value: Any,
) -> pa.Table:
    """
    Append a column to a ParquetFile by converting to PyArrow Table first.

    Args:
        parquet_file: The ParquetFile to add column to
        column_name: Name of the new column
        column_value: Value to populate in all rows of the new column

    Returns:
        PyArrow Table with the new column
    """
    # Convert ParquetFile to Table
    table = parquet_file.read()

    # Use the existing PyArrow append column function
    num_rows = table.num_rows
    column_array = pa.array([column_value] * num_rows)
    return table.append_column(column_name, column_array)


TABLE_CLASS_TO_APPEND_COLUMN_FUNC: Dict[
    Type[Union[LocalTable, DistributedDataset]], Callable
] = {
    pa.Table: pa_utils.append_column_to_table,
    papq.ParquetFile: append_column_to_parquet_file,
    pd.DataFrame: pd_utils.append_column_to_dataframe,
    pl.DataFrame: pl_utils.append_column_to_table,
    np.ndarray: np_utils.append_column_to_ndarray,
}

TABLE_CLASS_TO_SELECT_COLUMNS_FUNC: Dict[
    Type[Union[LocalTable, DistributedDataset]], Callable
] = {
    pa.Table: pa_utils.select_columns,
    pd.DataFrame: pd_utils.select_columns,
    pl.DataFrame: pl_utils.select_columns,
}

TABLE_CLASS_TO_TABLE_TYPE: Dict[Union[LocalTable, DistributedDataset], str] = {
    pa.Table: DatasetType.PYARROW.value,
    papq.ParquetFile: DatasetType.PYARROW_PARQUET.value,
    pl.DataFrame: DatasetType.POLARS.value,
    pd.DataFrame: DatasetType.PANDAS.value,
    np.ndarray: DatasetType.NUMPY.value,
    daft.DataFrame: DatasetType.DAFT.value,
    RayDataset: DatasetType.RAY_DATASET.value,
    MaterializedDataset: DatasetType.RAY_DATASET.value,
}

TABLE_TYPE_TO_DATASET_CREATE_FUNC: Dict[str, Callable] = {
    DatasetType.PYARROW.value: from_arrow,
    DatasetType.PYARROW_PARQUET.value: from_arrow,
    DatasetType.NUMPY.value: from_numpy,
    DatasetType.PANDAS.value: from_pandas,
}

TABLE_TYPE_TO_DATASET_CREATE_FUNC_REFS: Dict[str, Callable] = {
    DatasetType.PYARROW.value: from_arrow_refs,
    DatasetType.PYARROW_PARQUET.value: from_arrow_refs,
    DatasetType.NUMPY.value: from_numpy,
    DatasetType.PANDAS.value: from_pandas_refs,
    DatasetType.POLARS.value: from_arrow_refs,  # We cast Polars to Arrow for Ray Datasets
    DatasetType.RAY_DATASET.value: from_arrow_refs,  # Ray Datasets are created from Arrow refs
}

TABLE_TYPE_TO_CONCAT_FUNC: Dict[str, Callable] = {
    DatasetType.PYARROW_PARQUET.value: pa_utils.concat_tables,
    DatasetType.PYARROW.value: pa_utils.concat_tables,
    DatasetType.PANDAS.value: pd_utils.concat_dataframes,
    DatasetType.NUMPY.value: np_utils.concat_ndarrays,
    DatasetType.POLARS.value: pl_utils.concat_dataframes,
}


def _infer_schema_from_numpy_array(data: np.ndarray) -> Schema:
    """Infer schema from NumPy array."""
    if data.ndim > 2:
        raise ValueError(
            f"NumPy arrays with {data.ndim} dimensions are not supported. "
            f"Only 1D and 2D arrays are supported."
        )
    # Handle object dtype by converting to pandas first
    df = pd.DataFrame(data)
    arrow_schema = pa.Schema.from_pandas(df)

    from deltacat.storage.model.schema import Schema

    return Schema.of(schema=arrow_schema)


def _infer_schema_from_ray_dataset(data: RayDataset) -> Schema:
    """Infer schema from Ray Dataset."""
    ray_schema = data.schema()
    base_schema = ray_schema.base_schema

    if isinstance(base_schema, pa.Schema):
        arrow_schema = base_schema
    elif isinstance(base_schema, PandasBlockSchema):
        try:
            dtype_dict = {
                name: dtype for name, dtype in zip(base_schema.names, base_schema.types)
            }
            empty_df = pd.DataFrame(columns=base_schema.names).astype(dtype_dict)
            arrow_schema = pa.Schema.from_pandas(empty_df)
        except Exception as e:
            raise ValueError(
                f"Failed to convert Ray Dataset PandasBlockSchema to PyArrow schema: {e}"
            )
    else:
        raise ValueError(
            f"Unsupported Ray Dataset schema type: {type(base_schema)}. "
            f"Expected PyArrow Schema or PandasBlockSchema, got {base_schema}"
        )

    from deltacat.storage.model.schema import Schema

    return Schema.of(schema=arrow_schema)


def _infer_schema_from_pandas_dataframe(data: pd.DataFrame) -> Schema:
    """Infer schema from Pandas DataFrame."""
    from deltacat.storage.model.schema import Schema

    arrow_schema = pa.Schema.from_pandas(data)
    return Schema.of(schema=arrow_schema)


def _infer_schema_from_polars_dataframe(data: pl.DataFrame) -> Schema:
    """Infer schema from Polars DataFrame."""
    from deltacat.storage.model.schema import Schema

    arrow_table = data.to_arrow()
    return Schema.of(schema=arrow_table.schema)


def _infer_schema_from_pyarrow(
    data: Union[pa.Table, pa.RecordBatch, ds.Dataset]
) -> Schema:
    """Infer schema from PyArrow Table, RecordBatch, or Dataset."""
    from deltacat.storage.model.schema import Schema

    return Schema.of(schema=data.schema)


def _infer_schema_from_daft_dataframe(data: daft.DataFrame) -> Schema:
    """Infer schema from Daft DataFrame."""
    from deltacat.storage.model.schema import Schema

    daft_schema = data.schema()
    arrow_schema = daft_schema.to_pyarrow_schema()
    return Schema.of(schema=arrow_schema)


TABLE_CLASS_TO_SCHEMA_INFERENCE_FUNC: Dict[
    Type[Union[LocalTable, DistributedDataset], Callable]
] = {
    pd.DataFrame: _infer_schema_from_pandas_dataframe,
    pl.DataFrame: _infer_schema_from_polars_dataframe,
    pa.Table: _infer_schema_from_pyarrow,
    pa.RecordBatch: _infer_schema_from_pyarrow,
    ds.Dataset: _infer_schema_from_pyarrow,
    RayDataset: _infer_schema_from_ray_dataset,
    MaterializedDataset: _infer_schema_from_ray_dataset,  # MaterializedDataset uses same schema inference as RayDataset
    daft.DataFrame: _infer_schema_from_daft_dataframe,
    np.ndarray: _infer_schema_from_numpy_array,
}


def infer_table_schema(data: Union[LocalTable, DistributedDataset]) -> Schema:
    """Infer schema from a table or dataset."""
    infer_schema_func = _get_table_function(
        data,
        TABLE_CLASS_TO_SCHEMA_INFERENCE_FUNC,
        "schema inference",
    )
    return infer_schema_func(data)


def concat_tables(tables: List[LocalTable], table_type: DatasetType) -> LocalTable:
    """
    Concatenate a list of tables into a single table using the appropriate
    concatenation function for the given table type.

    Args:
        tables: List of tables to concatenate
        table_type: The DatasetType indicating which concatenation function to use

    Returns:
        Single concatenated table of the appropriate type

    Raises:
        ValueError: If no concatenation function is found for the table type
    """
    concat_func = _get_table_type_function(
        table_type, TABLE_TYPE_TO_CONCAT_FUNC, "concatenation"
    )
    return concat_func(tables)


def _daft_reader_wrapper(*args, **kwargs):
    """Wrapper for daft reader with lazy import to avoid circular import."""
    from deltacat.utils.daft import files_to_dataframe

    return files_to_dataframe(*args, **kwargs)


DISTRIBUTED_DATASET_TYPE_TO_READER_FUNC: Dict[int, Callable] = {
    DistributedDatasetType.DAFT.value: _daft_reader_wrapper,
}


class TableWriteMode(str, Enum):
    """
    Enum controlling how a given dataset will be written to a table.

    AUTO: CREATE if the table doesn't exist, APPEND if the table exists
    without merge keys, and MERGE if the table exists with merge keys.
    CREATE: Create the table if it doesn't exist, throw an error if it does.
    APPEND: Append to the table if it exists, throw an error if it doesn't.
    REPLACE: Replace existing table contents with the data to write.
    MERGE: Insert or update records matching table merge keys.
    Updates or inserts records based on the table's merge and sort keys by
    default.
    DELETE: Delete records matching table merge keys.
    """

    AUTO = "auto"
    CREATE = "create"
    APPEND = "append"
    REPLACE = "replace"
    MERGE = "merge"
    DELETE = "delete"


class SchemaEvolutionMode(str, Enum):
    """
    Enum controlling how schema changes are handled when writing to a table.

    MANUAL: Schema changes must be explicitly handled by the user. New fields
    not in the existing schema will cause an error.
    AUTO: Schema changes are automatically handled. New fields are added to
    the schema using the table's default_schema_consistency_type.
    DISABLED: Schema changes are disabled. The schema that the table was
    created with is immutable.
    """

    MANUAL = "manual"
    AUTO = "auto"
    DISABLED = "disabled"


class TableProperty(str, Enum):
    """
    Enum defining known table property key names.
    """

    READ_OPTIMIZATION_LEVEL = "read_optimization_level"
    RECORDS_PER_COMPACTED_FILE = "records_per_compacted_file"
    APPENDED_RECORD_COUNT_COMPACTION_TRIGGER = (
        "appended_record_count_compaction_trigger"
    )
    APPENDED_FILE_COUNT_COMPACTION_TRIGGER = "appended_file_count_compaction_trigger"
    APPENDED_DELTA_COUNT_COMPACTION_TRIGGER = "appended_delta_count_compaction_trigger"
    DEFAULT_COMPACTION_HASH_BUCKET_COUNT = "default_compaction_hash_bucket_count"
    SCHEMA_EVOLUTION_MODE = "schema_evolution_mode"
    DEFAULT_SCHEMA_CONSISTENCY_TYPE = "default_schema_consistency_type"
    SUPPORTED_READER_TYPES = "supported_reader_types"

    def read_table_property(
        table_or_table_version: Union[Table, TableVersion], property: TableProperty
    ) -> Any:
        properties = table_or_table_version.properties or {}
        value = properties.get(property.value, TablePropertyDefaultValues[property])

        # Handle property type conversion
        if property == TableProperty.SUPPORTED_READER_TYPES and isinstance(value, list):
            # Convert string values back to DatasetType enums
            return [DatasetType(v) for v in value]
        if property == TableProperty.SCHEMA_EVOLUTION_MODE:
            return SchemaEvolutionMode(value)
        if property == TableProperty.DEFAULT_SCHEMA_CONSISTENCY_TYPE:
            return SchemaConsistencyType(value)
        if property == TableProperty.READ_OPTIMIZATION_LEVEL:
            return TableReadOptimizationLevel(value)
        return value


class TableReadOptimizationLevel(str, Enum):
    """
    Enum controlling the how much to optimize reads when writing to a table. Different levels
    here correspond to different tradeoffs between write and read performance.

    NONE: No read optimization. Deletes and updates are resolved by finding the values
    that match merge key predicates by running compaction at read time. Provides the
    fastest/cheapest writes but slow/expensive reads. Resilient to conflicts with concurrent
    writes, including table management jobs like compaction.

    MODERATE: Discover record indexes that match merge key predicates at write time and record
    those values as logically deleted (e.g., using a bitmask). Provides faster/cheaper reads but
    slower/more-expensive writes. May conflict with concurrent writes that remove/replace data
    files like compaction.

    MAX: Materialize all deletes and updates at write time by running compaction during
    every write. Provides fast/cheap reads but slow/expensive writes. May conflict with
    concurrent writes, including table management jobs like compaction.
    """

    NONE = "none"
    MODERATE = "moderate"
    MAX = "max"


TablePropertyDefaultValues: Dict[TableProperty, Any] = {
    TableProperty.READ_OPTIMIZATION_LEVEL: TableReadOptimizationLevel.MAX,
    TableProperty.RECORDS_PER_COMPACTED_FILE: MAX_RECORDS_PER_COMPACTED_FILE,
    TableProperty.APPENDED_RECORD_COUNT_COMPACTION_TRIGGER: MAX_RECORDS_PER_COMPACTED_FILE
    * 2,
    TableProperty.APPENDED_FILE_COUNT_COMPACTION_TRIGGER: 1000,
    TableProperty.APPENDED_DELTA_COUNT_COMPACTION_TRIGGER: 100,
    TableProperty.DEFAULT_COMPACTION_HASH_BUCKET_COUNT: 8,
    TableProperty.SCHEMA_EVOLUTION_MODE: SchemaEvolutionMode.AUTO,
    TableProperty.DEFAULT_SCHEMA_CONSISTENCY_TYPE: SchemaConsistencyType.NONE,
    TableProperty.SUPPORTED_READER_TYPES: [d for d in DatasetType],
}


def _get_table_function(
    table: Union[LocalTable, DistributedDataset],
    function_map: Dict[Type, Callable],
    operation_name: str,
) -> Callable:
    """Generic helper to look up table-type-specific functions."""
    table_func = function_map.get(type(table))
    if table_func is None:
        msg = (
            f"No {operation_name} function found for table type: {type(table)}.\n"
            f"Known table types: {list(function_map.keys())}"
        )
        raise ValueError(msg)
    return table_func


def _get_table_type_function(
    table_type: DatasetType, function_map: Dict[str, Callable], operation_name: str
) -> Callable:
    """Generic helper to look up DatasetType-specific functions."""
    table_func = function_map.get(table_type.value)
    if table_func is None:
        msg = (
            f"No {operation_name} function found for table type: {table_type}.\n"
            f"Known table types: {list(function_map.keys())}"
        )
        raise ValueError(msg)
    return table_func


def _convert_all(tables: List[LocalTable], conversion_fn: Callable, **kwargs):
    if not tables:  # Empty list
        return pd.DataFrame()

    # Convert list elements to the same type
    all_tables = []
    for i, table in enumerate(tables):
        try:
            converted_table = conversion_fn(table, **kwargs)
            all_tables.append(converted_table)
        except Exception as e:
            raise ValueError(f"Failed to convert list element {i}: {e}") from e

    # Concatenate with error handling
    try:
        return concat_tables(all_tables, get_dataset_type(all_tables[0]))
    except Exception as e:
        raise ValueError(f"Failed to concatenate {len(all_tables)} tables: {e}") from e


def get_table_length(
    table: Union[LocalTable, DistributedDataset, BlockAccessor]
) -> int:
    """
    Generic function to get the length of a table or distributed dataset.

    Args:
        table: The local table or distributed dataset to get the length of

    Returns:
        Length of the table or distributed dataset in rows
    """
    # Handle DAFT DataFrames dynamically
    if hasattr(table, "count_rows") and str(type(table).__module__).startswith("daft"):
        return table.count_rows()
    elif isinstance(table, RayDataset):
        return table.count()
    elif isinstance(table, papq.ParquetFile):
        return table.metadata.num_rows
    else:
        return len(table)


def dataset_length(table: Dataset) -> int:
    """
    Generic function to get the length of a dataset in records.
    If the input is a list of tables, the length is the sum of the
    lengths of the tables.

    Args:
        table: The dataset to get the length of

    Returns:
        Length of the dataset in records
    """
    if isinstance(table, list):
        return sum(get_table_length(t) for t in table)
    return get_table_length(table)


def get_table_size(table: Union[LocalTable, DistributedDataset]) -> int:
    """
    Generic function to get the size of a table or distributed dataset.

    Args:
        table: The local table or distributed dataset to get the size of

    Returns:
        Size of the table or distributed dataset
    """
    table_size_func = _get_table_function(table, TABLE_CLASS_TO_SIZE_FUNC, "size")
    return table_size_func(table)


def dataset_size(table: Dataset) -> int:
    """
    Generic function to get the size of a dataset in bytes.
    If the input is a list of tables, the size is the sum of the
    sizes of the tables.

    Args:
        table: The dataset to get the size of

    Returns:
        Size of the dataset in bytes
    """
    if isinstance(table, list):
        return sum(get_table_size(t) for t in table)
    return get_table_size(table)


def get_table_column_names(table: Union[LocalTable, DistributedDataset]) -> List[str]:
    """
    Generic function to get the column names of a table or distributed dataset.

    Args:
        table: The local table or distributed dataset to get the column names of

    Returns:
        List of column names
    """
    column_names_func = _get_table_function(
        table, TABLE_CLASS_TO_COLUMN_NAMES_FUNC, "column names"
    )
    return column_names_func(table)


def dataset_column_names(table: Dataset) -> List[str]:
    """
    Generic function to get the column names of a dataset.
    If the input is a list of tables, unique column names are
    returned in the order they are first seen in the list.

    Args:
        table: The dataset to get the column names of

    Returns:
        List of column names
    """
    if isinstance(table, list):
        # use dictionary keys as an ordered set
        column_names = {}
        for t in table:
            for column_name in get_table_column_names(t):
                column_names[column_name] = None
        return list(column_names.keys())
    return get_table_column_names(table)


def get_table_schema(table: Union[LocalTable, DistributedDataset]) -> pa.Schema:
    """
    Generic function to get the PyArrow schema of a table or distributed dataset.

    Args:
        table: The local table or distributed dataset to get the schema of

    Returns:
        PyArrow Schema object
    """
    schema_func = _get_table_function(table, TABLE_CLASS_TO_SCHEMA_FUNC, "schema")
    return schema_func(table)


def dataset_schema(table: Dataset) -> pa.Schema:
    """
    Generic function to get the PyArrow schema of a dataset. If the input is a list of
    tables, uses pyarrow.unify_schemas(schemas, promote_options="permissive").

    Args:
        table: The dataset to get the schema of

    Returns:
        PyArrow Schema object
    """
    if isinstance(table, list):
        return pa.unify_schemas(
            [get_table_schema(t) for t in table], promote_options="permissive"
        )
    return get_table_schema(table)


def get_table_writer(table: Union[LocalTable, DistributedDataset]) -> Callable:
    """
    Generic function to get a table writer function for a given dataset type.

    Args:
        table: The local table or distributed dataset to get the writer function for

    Returns:
        Writer function for the given dataset type
    """
    return _get_table_function(table, TABLE_CLASS_TO_WRITER_FUNC, "writer")


def get_table_slicer(table: Union[LocalTable, DistributedDataset]) -> Callable:
    """
    Generic function to get a table slicer function for a given dataset type.

    Args:
        table: The local table or distributed dataset to get the slicer function for

    Returns:
        Slicer function for the given dataset type
    """
    return _get_table_function(table, TABLE_CLASS_TO_SLICER_FUNC, "slicer")


def get_dataset_type(dataset: Union[LocalTable, DistributedDataset]) -> DatasetType:
    """Get the DatasetType enum value for a given dataset object.

    Args:
        dataset: The dataset object to identify

    Returns:
        DatasetType enum value corresponding to the dataset type

    Raises:
        ValueError: If the dataset type is not supported
    """
    dataset_type_str = _get_table_function(
        dataset, TABLE_CLASS_TO_TABLE_TYPE, "dataset type identification"
    )
    return DatasetType(dataset_type_str)


def table_to_pyarrow(
    table: Union[LocalTable, DistributedDataset],
    *,
    schema: Optional[pa.Schema] = None,
    **kwargs,
) -> pa.Table:
    """
    Convert a single table or distributed dataset to PyArrow Table format.

    Args:
        table: The local table or distributed dataset to convert
        schema: Optional schema to use for the conversion
        **kwargs: Additional arguments passed to the conversion function

    Returns:
        PyArrow Table created from the provided dataset
    """
    to_pyarrow_func = _get_table_function(
        table, TABLE_CLASS_TO_PYARROW_FUNC, "pyarrow conversion"
    )
    return to_pyarrow_func(table, schema=schema, **kwargs)


def table_to_pandas(
    table: Union[LocalTable, DistributedDataset],
    *,
    schema: Optional[pa.Schema] = None,
    **kwargs,
) -> pd.DataFrame:
    """
    Convert a single table or distributed dataset to pandas DataFrame format.

    Args:
        table: The local table or distributed dataset to convert
        schema: Optional schema to use for the conversion
        **kwargs: Additional arguments passed to the conversion function

    Returns:
        pandas DataFrame created from the provided dataset
    """
    to_pandas_func = _get_table_function(
        table, TABLE_CLASS_TO_PANDAS_FUNC, "pandas conversion"
    )
    return to_pandas_func(table, schema=schema, **kwargs)


def to_pyarrow(
    table: Dataset, *, schema: Optional[pa.Schema] = None, **kwargs
) -> pa.Table:
    """
    Convert any supported dataset type to PyArrow Table format.

    Args:
        table: The table/dataset to convert
        schema: Optional schema to use for the conversion
        **kwargs: Additional arguments passed to the conversion function

    Returns:
        PyArrow Table created from the provided dataset
    """
    if isinstance(table, list):
        return _convert_all(table, table_to_pyarrow, schema=schema, **kwargs)
    return table_to_pyarrow(table, schema=schema, **kwargs)


def to_pandas(
    table: Dataset, *, schema: Optional[pa.Schema] = None, **kwargs
) -> pd.DataFrame:
    """
    Convert any supported dataset type to pandas DataFrame format.

    Args:
        table: The table/dataset to convert
        schema: Optional schema to use for the conversion
        **kwargs: Additional arguments passed to the conversion function

    Returns:
        pandas DataFrame created from the provided dataset
    """
    if isinstance(table, list):
        return _convert_all(table, table_to_pandas, schema=schema, **kwargs)
    return table_to_pandas(table, schema=schema, **kwargs)


def from_pyarrow(pa_table: pa.Table, target_type: DatasetType, **kwargs) -> Dataset:
    """Convert PyArrow Table to the specified dataset type.

    Args:
        pa_table: PyArrow Table to convert
        target_type: Target DatasetType to convert to
        **kwargs: Additional arguments passed to the conversion function

    Returns:
        Dataset converted to the target type

    Raises:
        ValueError: If target_type is not supported
    """
    conversion_func = _get_table_type_function(
        target_type,
        DATASET_TYPE_FROM_PYARROW,
        f"{target_type} conversion",
    )
    return conversion_func(pa_table, **kwargs)


def from_pandas(pd_df: pd.DataFrame, target_type: DatasetType, **kwargs) -> Dataset:
    """Convert Pandas DataFrame to the specified dataset type.

    Args:
        pd_df: Pandas DataFrame to convert
        target_type: Target DatasetType to convert to
        **kwargs: Additional arguments passed to the conversion function

    Returns:
        Dataset converted to the target type

    Raises:
        ValueError: If target_type is not supported
    """
    conversion_func = _get_table_type_function(
        target_type,
        DATASET_TYPE_FROM_PANDAS,
        f"{target_type} conversion",
    )
    return conversion_func(pd_df, **kwargs)


def empty_table(table_type: DatasetType) -> Dataset:
    """
    Create an empty table of the given type.
    """
    empty_table_func = _get_table_type_function(
        table_type, TABLE_TYPE_TO_EMPTY_TABLE_FUNC, "empty table"
    )
    return empty_table_func()


def append_column_to_table(
    table: LocalTable,
    column_name: str,
    column_value: Any,
) -> LocalTable:
    """
    Generic function to append a column with a specified value to any supported dataset type.

    Args:
        table: The table/dataset to add column to
        column_name: Name of the new column
        column_value: Value to populate in all rows of the new column
        table_type: Type of the dataset

    Returns:
        Updated table with the new column
    """
    append_column_to_table_func = _get_table_function(
        table, TABLE_CLASS_TO_APPEND_COLUMN_FUNC, "append column"
    )
    return append_column_to_table_func(table, column_name, column_value)


def select_columns_from_table(
    table: LocalTable,
    column_names: List[str],
) -> LocalTable:
    """
    Generic function to select columns from any supported dataset type.

    Args:
        table: The table/dataset to select columns from
        column_names: List of column names to select

    Returns:
        Updated table with the selected columns
    """
    select_columns_func = _get_table_function(
        table, TABLE_CLASS_TO_SELECT_COLUMNS_FUNC, "select columns"
    )
    return select_columns_func(table, column_names)


def write_sliced_table(
    table: Union[LocalTable, DistributedDataset],
    base_path: str,
    filesystem: Optional[pa.fs.FileSystem],
    max_records_per_entry: Optional[int],
    table_writer_fn: Callable,
    table_slicer_fn: Callable,
    table_writer_kwargs: Optional[Dict[str, Any]] = None,
    content_type: ContentType = ContentType.PARQUET,
    entry_params: Optional[EntryParams] = None,
    entry_type: Optional[EntryType] = EntryType.DATA,
) -> ManifestEntryList:
    """
    Writes table slices to 1 or more files and returns
    manifest entries describing the uploaded files.

    Args:
        table: The local table or distributed dataset to write
        base_path: The base path to write the table to
        filesystem: The filesystem to write the table to
        table_writer_fn: The function to write the table to
        table_slicer_fn: The function to slice the table into multiple files
        table_writer_kwargs: Additional arguments to pass to the table writer
        content_type: The content type to write the table to
        entry_params: Manifest entry parameters
        entry_type: The manifest entry types to write

    Returns:
        Manifest entries describing the uploaded files
    """
    # @retry decorator can't be pickled by Ray, so wrap upload in Retrying
    retrying = Retrying(
        wait=wait_random_exponential(multiplier=1, max=60),
        stop=stop_after_delay(UPLOAD_SLICED_TABLE_RETRY_STOP_AFTER_DELAY),
        retry=retry_if_exception_type(RetryableError),
    )

    manifest_entries = ManifestEntryList()
    table_record_count = get_table_length(table)

    if max_records_per_entry is None or not table_record_count:
        # write the whole table to a single file
        manifest_entries = retrying(
            write_table,
            table,
            f"{base_path}",  # cast any non-string arg to string
            filesystem,
            table_writer_fn,
            table_writer_kwargs,
            content_type,
            entry_params,
            entry_type,
        )
    else:
        # iteratively write table slices
        table_slices = table_slicer_fn(table, max_records_per_entry)
        for table_slice in table_slices:
            slice_entries = retrying(
                write_table,
                table_slice,
                f"{base_path}",  # cast any non-string arg to string
                filesystem,
                table_writer_fn,
                table_writer_kwargs,
                content_type,
                entry_params,
                entry_type,
            )
            manifest_entries.extend(slice_entries)
    return manifest_entries


def write_table(
    table: Union[LocalTable, DistributedDataset],
    base_path: str,
    filesystem: Optional[pa.fs.FileSystem],
    table_writer_fn: Callable,
    table_writer_kwargs: Optional[Dict[str, Any]],
    content_type: ContentType = ContentType.PARQUET,
    entry_params: Optional[EntryParams] = None,
    entry_type: Optional[EntryType] = EntryType.DATA,
) -> ManifestEntryList:
    """
    Writes the given table to 1 or more files and return
    manifest entries describing the uploaded files.

    Args:
        table: The local table or distributed dataset to write
        base_path: The base path to write the table to
        filesystem: The filesystem to write the table to
        table_writer_fn: The function to write the table to
        table_writer_kwargs: Additional arguments to pass to the table writer
        content_type: The content type to write the table to
        entry_params: Manifest entry parameters
        entry_type: The manifest entry types to write

    Returns:
        Manifest entries describing the uploaded files
    """
    if table_writer_kwargs is None:
        table_writer_kwargs = {}

    # Determine content_encoding before writing files so we can include it in filenames
    content_encoding = None
    if content_type in EXPLICIT_COMPRESSION_CONTENT_TYPES:
        # TODO(pdames): Support other user-specified encodings at write time.
        content_encoding = ContentEncoding.GZIP

    wrapped_obj = (
        CapturedBlockWritePathsActor.remote()
        if isinstance(table, RayDataset)
        else CapturedBlockWritePathsBase()
    )
    capture_object = CapturedBlockWritePaths(wrapped_obj)
    block_write_path_provider = UuidBlockWritePathProvider(
        capture_object,
        base_path=base_path,
        content_type=content_type,
        content_encoding=content_encoding,
    )
    # Extract schema, schema_id, and sort_scheme_id from table_writer_kwargs
    schema = table_writer_kwargs.pop("schema", None)
    schema_id = table_writer_kwargs.pop("schema_id", None)
    sort_scheme_id = table_writer_kwargs.pop("sort_scheme_id", None)
    table_writer_fn(
        table,
        base_path,
        filesystem,
        block_write_path_provider,
        content_type.value,
        schema=schema,
        **table_writer_kwargs,
    )
    # TODO: Add a proper fix for block_refs and write_paths not persisting in Ray actors
    del block_write_path_provider
    blocks = capture_object.blocks()
    write_paths = capture_object.write_paths()
    metadata = get_block_metadata_list(table, write_paths, blocks)
    manifest_entries = ManifestEntryList()
    for block_idx, path in enumerate(write_paths):
        try:
            manifest_entry = ManifestEntry.from_path(
                path=path,
                filesystem=filesystem,
                record_count=metadata[block_idx].num_rows,
                source_content_length=metadata[block_idx].size_bytes,
                content_type=content_type.value,
                content_encoding=content_encoding,
                entry_type=entry_type,
                entry_params=entry_params,
                schema_id=schema_id,
                sort_scheme_id=sort_scheme_id,
            )
            manifest_entries.append(manifest_entry)
        except RETRYABLE_TRANSIENT_ERRORS as e:
            _handle_retryable_error(e, path, "write", RetryableUploadTableError)
        except BaseException as e:
            _handle_non_retryable_error(
                e,
                path,
                "upload",
                NonRetryableUploadTableError,
                f"and content_type={content_type}",
            )
    return manifest_entries


@ray.remote
class CapturedBlockWritePathsActor:
    def __init__(self):
        self._wrapped = CapturedBlockWritePathsBase()

    def extend(self, write_paths: List[str], blocks: List[Block]) -> None:
        self._wrapped.extend(write_paths, blocks)

    def write_paths(self) -> List[str]:
        return self._wrapped.write_paths()

    def blocks(self) -> List[Block]:
        return self._wrapped.blocks()


class CapturedBlockWritePathsBase:
    def __init__(self):
        self._write_paths: List[str] = []
        self._blocks: List[Block] = []

    def extend(self, write_paths: List[str], blocks: List[Block]) -> None:
        try:
            iter(write_paths)
        except TypeError:
            pass
        else:
            self._write_paths.extend(write_paths)
        try:
            iter(blocks)
        except TypeError:
            pass
        else:
            self._blocks.extend(blocks)

    def write_paths(self) -> List[str]:
        return self._write_paths

    def blocks(self) -> List[Block]:
        return self._blocks


class CapturedBlockWritePaths:
    def __init__(self, wrapped=CapturedBlockWritePathsBase()):
        self._wrapped = wrapped

    def extend(self, write_paths: List[str], blocks: List[Block]) -> None:
        return (
            self._wrapped.extend(write_paths, blocks)
            if isinstance(self._wrapped, CapturedBlockWritePathsBase)
            else ray.get(self._wrapped.extend.remote(write_paths, blocks))
        )

    def write_paths(self) -> List[str]:
        return (
            self._wrapped.write_paths()
            if isinstance(self._wrapped, CapturedBlockWritePathsBase)
            else ray.get(self._wrapped.write_paths.remote())
        )

    def blocks(self) -> List[Block]:
        return (
            self._wrapped.blocks()
            if isinstance(self._wrapped, CapturedBlockWritePathsBase)
            else ray.get(self._wrapped.blocks.remote())
        )


class UuidBlockWritePathProvider(FilenameProvider):
    """Block write path provider implementation that writes each
    dataset block out to a file of the form: {base_path}/{uuid}
    """

    def __init__(
        self,
        capture_object: CapturedBlockWritePaths,
        base_path: Optional[str] = None,
        content_type: Optional[ContentType] = None,
        content_encoding: Optional[ContentEncoding] = None,
    ):
        self.base_path = base_path
        self.content_type = content_type
        self.content_encoding = content_encoding
        self.write_paths: List[str] = []
        self.blocks: List[Block] = []
        self.capture_object = capture_object

    def __del__(self):
        if self.write_paths or self.blocks:
            self.capture_object.extend(
                self.write_paths,
                self.blocks,
            )

    def get_filename_for_block(
        self,
        block: Block,
        task_index: int,
        block_index: int,
    ) -> str:
        if self.base_path is None:
            raise ValueError(
                "Base path must be provided to UuidBlockWritePathProvider",
            )
        return self._get_write_path_for_block(
            base_path=self.base_path,
            block=block,
            block_index=block_index,
        )

    def _get_write_path_for_block(
        self,
        base_path: str,
        *,
        block: Optional[Block] = None,
        **kwargs,
    ) -> str:
        # Generate base UUID filename
        filename = str(uuid4())

        # Add content type extension if available
        if self.content_type:
            content_type_extension = None
            content_type_extension = CONTENT_TYPE_TO_EXT.get(self.content_type)
            if content_type_extension:
                filename += content_type_extension

        # Add content encoding extension if available
        if self.content_encoding:
            encoding_extension = None
            encoding_extension = CONTENT_ENCODING_TO_EXT.get(self.content_encoding)
            if encoding_extension:
                filename += encoding_extension

        write_path = f"{base_path}/{filename}"
        self.write_paths.append(write_path)
        if block is not None:
            self.blocks.append(block)
        return filename

    def __call__(
        self,
        base_path: str,
        *,
        filesystem: Optional[pyarrow.fs.FileSystem] = None,
        dataset_uuid: Optional[str] = None,
        block: Optional[Block] = None,
        block_index: Optional[int] = None,
        file_format: Optional[str] = None,
    ) -> str:
        return self._get_write_path_for_block(
            base_path,
            filesystem=filesystem,
            dataset_uuid=dataset_uuid,
            block=block,
            block_index=block_index,
            file_format=file_format,
        )


def get_block_metadata_list(
    table: LocalTable,
    write_paths: List[str],
    blocks: List[Block],
) -> List[BlockMetadata]:
    """
    Get the block metadata for a given table.

    Args:
        table: The local table or distributed dataset to get the block metadata for
        write_paths: The list of write paths for the table
        blocks: The list of blocks to get the metadata for

    Returns:
        List of block metadata
    """
    block_meta_list: List[BlockMetadata] = []
    if not blocks:
        # this must be a local table - ensure it was written to only 1 file
        assert len(write_paths) == 1, (
            f"Expected table of type '{type(table)}' to be written to 1 "
            f"file, but found {len(write_paths)} files."
        )
        blocks = [table]
    for block in blocks:
        block_meta_list.append(get_block_metadata(block))
    return block_meta_list


def get_block_metadata(
    table: Union[LocalTable, DistributedDataset, BlockAccessor],
) -> BlockMetadata:
    """
    Get the block metadata for a given table.

    Args:
        table: The local table or distributed dataset to get the block metadata for

    Returns:
        Block metadata
    """
    table_size = None
    table_size_func = TABLE_CLASS_TO_SIZE_FUNC.get(type(table))
    if table_size_func:
        table_size = table_size_func(table)
    else:
        logger.warning(f"Unable to estimate '{type(table)}' table size.")
    if isinstance(table, BlockAccessor):
        table = table.to_block()
    return BlockMetadata(
        num_rows=get_table_length(table),
        size_bytes=table_size,
        schema=None,
        input_files=None,
        exec_stats=None,
    )


def _reconstruct_manifest_entry_uri(
    manifest_entry: ManifestEntry,
    **kwargs,
) -> ManifestEntry:
    """
    Reconstruct the full URI for a manifest entry.

    Args:
        manifest_entry: The manifest entry to reconstruct the URI for
        **kwargs: Additional arguments to pass to the catalog properties

    Returns:
        Manifest entry with the reconstructed URI
    """
    # Reconstruct full URI with scheme for external readers (see GitHub issue #567)
    from deltacat.catalog import get_catalog_properties

    # Only pass kwargs that CatalogProperties actually accepts
    catalog_kwargs = _filter_kwargs_for_catalog_properties(kwargs)
    catalog_properties = get_catalog_properties(**catalog_kwargs)

    original_uri = manifest_entry.uri
    reconstructed_uri = catalog_properties.reconstruct_full_path(original_uri)
    if original_uri != reconstructed_uri:
        # Create a copy of the manifest entry with the reconstructed URI
        reconstructed_entry = ManifestEntry(
            uri=reconstructed_uri, url=manifest_entry.url, meta=manifest_entry.meta
        )
        return reconstructed_entry
    return manifest_entry


def _filter_kwargs_for_external_readers(kwargs: Dict[str, Any]) -> Dict[str, Any]:
    """
    Filter out DeltaCAT system kwargs that external file readers don't expect.

    Use this when passing kwargs to external libraries like PyArrow, Pandas, Polars, etc.
    This removes all DeltaCAT-specific parameters that would cause TypeErrors in external readers.

    Args:
        kwargs: The dictionary of arguments to filter

    Returns:
        Dictionary of arguments with DeltaCAT system kwargs removed
    """
    return {
        k: v
        for k, v in kwargs.items()
        if k
        not in [
            # DeltaCAT catalog/storage system kwargs
            "inner",
            "catalog",
            "ray_options_provider",
            "distributed_dataset_type",
            # DeltaCAT schema/reader kwargs
            "table_version_schema",
            "entry_params",
            # Daft-specific kwargs
            "io_config",
            "ray_init_options",
            # DeltaCAT processing kwargs
            "column_names",
            "include_columns",
            "file_reader_kwargs_provider",
            "file_path_column",
            "max_parallelism",
        ]
    }


def _filter_kwargs_for_catalog_properties(kwargs: Dict[str, Any]) -> Dict[str, Any]:
    """
    Filter kwargs to only include those that CatalogProperties accepts.

    Use this when calling get_catalog_properties() or CatalogProperties.__init__().
    Uses a whitelist approach - only passes known compatible parameters.

    CatalogProperties.__init__ accepts: root, filesystem, storage
    get_catalog_properties also accepts: catalog, inner

    Args:
        kwargs: The dictionary of arguments to filter

    Returns:
        Dictionary containing only CatalogProperties-compatible kwargs
    """
    return {
        k: v
        for k, v in kwargs.items()
        if k in ["root", "filesystem", "storage", "catalog", "inner"]
    }


def _filter_kwargs_for_reader_functions(kwargs: Dict[str, Any]) -> Dict[str, Any]:
    """
    Filter kwargs for internal DeltaCAT reader functions that need most params but not catalog-specific ones.

    Use this for internal DeltaCAT functions that need file reader kwargs, schema kwargs, etc.
    but should not receive catalog/storage system parameters.
    Preserves table_version_schema, entry_params, file reader kwargs, etc.

    Args:
        kwargs: The dictionary of arguments to filter

    Returns:
        Dictionary with catalog/storage system kwargs removed
    """
    return {
        k: v
        for k, v in kwargs.items()
        if k
        not in ["inner", "catalog", "ray_options_provider", "distributed_dataset_type"]
    }


def _extract_content_metadata(
    manifest_entry: ManifestEntry,
) -> Tuple[ContentType, ContentEncoding, str]:
    """
    Extract content type, encoding, and path from manifest entry.

    Args:
        manifest_entry: The manifest entry to extract the content metadata from

    Returns:
        Tuple of content type, encoding, and path
    """
    content_type = manifest_entry.meta.content_type
    assert content_type, f"Unknown content type for manifest entry: {manifest_entry}"
    content_type = ContentType(content_type)

    content_encoding = manifest_entry.meta.content_encoding
    assert (
        content_encoding
    ), f"Unknown content encoding for manifest entry: {manifest_entry}"
    content_encoding = ContentEncoding(content_encoding)

    path = manifest_entry.uri
    if path is None:
        path = manifest_entry.url

    return content_type, content_encoding, path


def _extract_partial_download_params(
    manifest_entry: ManifestEntry,
) -> Optional[PartialFileDownloadParams]:
    """
    Extract partial file download parameters from manifest entry.

    Args:
        manifest_entry: The manifest entry to extract the partial file download parameters from

    Returns:
        Partial file download parameters
    """
    if not manifest_entry.meta or not manifest_entry.meta.content_type_parameters:
        return None

    for type_params in manifest_entry.meta.content_type_parameters:
        if isinstance(type_params, PartialFileDownloadParams):
            return type_params
    return None


def _create_retry_wrapper():
    """
    Create a standardized Tenacity Retrying wrapper for file operations.

    Returns:
        Tenacity Retrying wrapper
    """
    return Retrying(
        wait=wait_random_exponential(multiplier=1, max=60),
        stop=stop_after_delay(DOWNLOAD_MANIFEST_ENTRY_RETRY_STOP_AFTER_DELAY),
        retry=retry_if_exception_type(RetryableError),
    )


def _remove_file_path_column(
    include_columns: Optional[List[str]],
    file_path_column: Optional[str],
) -> Optional[List[str]]:
    """Remove the file path system column from the include_columns list.

    Args:
        include_columns: The list of columns to include in a selection
        file_path_column: Optional file path system column name to remove from the selection

    Returns:
        List of columns to include without the file path system column
    """
    if file_path_column and include_columns:
        return [col for col in include_columns if col != file_path_column]
    return include_columns


def _prepare_download_arguments(
    table_type: DatasetType,
    column_names: Optional[List[str]],
    include_columns: Optional[List[str]],
    file_reader_kwargs_provider: Optional[ReadKwargsProvider],
    file_path_column: Optional[str],
    **kwargs,
) -> Dict[str, Any]:
    """Prepare standardized arguments for download operations.

    Args:
        table_type: The type of table to download
        column_names: The list of column names in the table
        include_columns: The list of columns to include in the selection
        file_reader_kwargs_provider: Optional per-content-type provider of file reader kwargs
        file_path_column: The file path system column name
        **kwargs: Additional arguments to pass to the file reader

    Returns:
        Dictionary of arguments for the download operation
    """
    reader_kwargs = _filter_kwargs_for_external_readers(kwargs)
    processed_include_columns = _remove_file_path_column(
        include_columns, file_path_column
    )

    return {
        "table_type": table_type,
        "column_names": column_names,
        "include_columns": processed_include_columns,
        "file_reader_kwargs_provider": file_reader_kwargs_provider,
        "file_path_column": file_path_column,
        **reader_kwargs,
    }


def _handle_retryable_error(e: Exception, path: str, operation: str, error_class: type):
    """Handle retryable errors with standardized error message."""
    raise error_class(
        f"Retry {operation} for: {path} after receiving {type(e).__name__}: {e}"
    ) from e


def _handle_non_retryable_error(
    e: Exception, path: str, operation: str, error_class: type, extra_context: str = ""
):
    """Handle non-retryable errors with logging and standardized error message."""
    context = f" {extra_context}" if extra_context else ""
    logger.warning(
        f"{operation.title()} has failed for {path}{context}. Error: {e}",
        exc_info=True,
    )
    raise error_class(
        f"{operation.title()} has failed for {path}{context}: Error: {e}"
    ) from e


def from_manifest_table(
    manifest_table: Union[LocalDataset, DistributedDataset],
    dataset_type: DatasetType = DatasetType.DAFT,
    schema: Optional[pa.Schema] = None,
    **kwargs,
) -> Dataset:
    """
    Read a manifest table (containing file paths and metadata) and download the actual data.

    This utility function takes the output from a schemaless table read (which returns
    manifest entries instead of data) and downloads the actual file contents.

    Args:
        manifest_table: Dataset containing manifest entries with file paths and metadata
        dataset_type: The type of dataset to return (DAFT, RAY_DATASET, PYARROW, etc.)
        schema: Optional PyArrow schema to enforce consistent column names across formats
        **kwargs: Additional arguments forwarded to download functions

    Returns:
        Dataset containing the actual file contents
    """
    # Convert the manifest table to pandas for easier processing
    # TODO(pdames): Iterate over each input manifest table in its native format
    manifest_df = to_pandas(manifest_table)

    # Reconstruct ManifestEntry objects from the manifest data
    manifest_entries = []
    for _, row in manifest_df.iterrows():
        # Create ManifestMeta from the row data
        meta = ManifestMeta.of(
            content_length=row.get("meta_content_length"),
            record_count=row.get("meta_record_count"),
            content_type=row.get("meta_content_type"),
            content_encoding=row.get("meta_content_encoding"),
        )

        # Create ManifestEntry
        entry = ManifestEntry.of(
            url=row["path"],
            meta=meta,
            mandatory=row.get("mandatory", True),
            uuid=row.get("id"),
        )
        manifest_entries.append(entry)

    # Create a new Manifest from the entries
    reconstructed_manifest = Manifest.of(entries=manifest_entries)

    # Add schema to kwargs if provided
    if schema is not None:
        kwargs["table_version_schema"] = schema

    # Choose the appropriate download function based on dataset type
    if dataset_type in DatasetType.distributed():
        # Use distributed download function
        # Map DatasetType to DistributedDatasetType
        distributed_type_map = {
            DatasetType.DAFT: DistributedDatasetType.DAFT,
            DatasetType.RAY_DATASET: DistributedDatasetType.RAY_DATASET,
        }
        distributed_dataset_type = distributed_type_map.get(dataset_type)
        if distributed_dataset_type is None:
            raise ValueError(f"Unsupported distributed dataset type: {dataset_type}")

        return download_manifest_entries_distributed(
            manifest=reconstructed_manifest,
            distributed_dataset_type=distributed_dataset_type,
            **kwargs,
        )
    else:
        # Use local download function
        return download_manifest_entries(
            manifest=reconstructed_manifest,
            table_type=dataset_type,
            **kwargs,
        )


def download_manifest_entries(
    manifest: Manifest,
    table_type: DatasetType = DatasetType.PYARROW,
    max_parallelism: Optional[int] = 1,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
    file_path_column: Optional[str] = None,
    **kwargs,
) -> LocalDataset:
    """Download all entries in the manifest.

    Args:
        manifest: The manifest containing the entries to download
        table_type: Dataset type to load the entries into
        max_parallelism: Maximum parallelism to use
        column_names: The list of column names in the table
        include_columns: The list of columns to include in the selection
        file_reader_kwargs_provider: Optional per-content-type provider of file reader kwargs,
            (e.g., to pass in a custom schema for a Parquet file)
        file_path_column: Optional file path system column name
        **kwargs: Additional arguments to pass to the file reader

    Returns:
        Local dataset
    """
    if max_parallelism and max_parallelism <= 1:
        return _download_manifest_entries(
            manifest,
            table_type,
            column_names,
            include_columns,
            file_reader_kwargs_provider,
            file_path_column,
            **kwargs,
        )
    else:
        return _download_manifest_entries_parallel(
            manifest,
            table_type,
            max_parallelism,
            column_names,
            include_columns,
            file_reader_kwargs_provider,
            file_path_column,
            **kwargs,
        )


def _download_manifest_entries(
    manifest: Manifest,
    table_type: DatasetType = DatasetType.PYARROW,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
    file_path_column: Optional[str] = None,
    **kwargs,
) -> LocalDataset:
    """Download all entries in the manifest.

    Args:
        manifest: The manifest containing the entries to download
        table_type: Dataset type to load the entries into
        column_names: The list of column names in the table
        include_columns: The list of columns to include in the selection
        file_reader_kwargs_provider: Optional per-content-type provider of file reader kwargs,
            (e.g., to pass in a custom schema for a Parquet file)
        file_path_column: Optional file path system column name
        **kwargs: Additional arguments to pass to the file reader

    Returns:
        Local dataset
    """
    download_args = _prepare_download_arguments(
        table_type,
        column_names,
        include_columns,
        file_reader_kwargs_provider,
        file_path_column,
        **kwargs,
    )
    result = []
    for e in manifest.entries:
        manifest_entry = _reconstruct_manifest_entry_uri(e, **kwargs)
        result.append(
            download_manifest_entry(manifest_entry=manifest_entry, **download_args)
        )

    return result


@ray.remote
def download_manifest_entry_ray(
    manifest_entry: ManifestEntry,
    table_type: DatasetType = DatasetType.PYARROW,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
    content_type: Optional[ContentType] = None,
    content_encoding: Optional[ContentEncoding] = None,
    filesystem: Optional[pyarrow.fs.FileSystem] = None,
    file_path_column: Optional[str] = None,
    **kwargs,
) -> LocalTable:
    """
    Ray remote function for downloading manifest entries.

    Args:
        manifest_entry: The manifest entry to download
        table_type: Dataset type to load the entry into
        column_names: The list of column names in the table
        include_columns: The list of columns to include in the selection
        file_reader_kwargs_provider: Optional per-content-type provider of file reader kwargs,
            (e.g., to pass in a custom schema for a Parquet file)
        content_type: Optional content type of the file
        content_encoding: Optional content encoding of the file
        filesystem: Optional PyArrow filesystem to use to read the file
        file_path_column: Optional file path system column name
        **kwargs: Additional arguments to pass to the file reader

    Returns:
        Local table
    """
    # Make sure we normalize the table type to PyArrow to provide the correct
    # input type to from_arrow_refs
    effective_table_type = table_type
    if table_type == DatasetType.RAY_DATASET:
        effective_table_type = DatasetType.PYARROW

    # Call the regular download function
    result = download_manifest_entry(
        manifest_entry=manifest_entry,
        table_type=effective_table_type,
        column_names=column_names,
        include_columns=include_columns,
        file_reader_kwargs_provider=file_reader_kwargs_provider,
        content_type=content_type,
        content_encoding=content_encoding,
        filesystem=filesystem,
        file_path_column=file_path_column,
        **kwargs,
    )

    # Convert Polars DataFrame to Arrow Table for Ray dataset compatibility
    if isinstance(result, pl.DataFrame):
        result = result.to_arrow()

    # Cast string_view columns to string to avoid cloudpickle issues
    if isinstance(result, pa.Table):
        result = _cast_string_view_to_string(result)

    return result


def download_manifest_entries_distributed(
    manifest: Manifest,
    table_type: DatasetType = DatasetType.PYARROW,
    max_parallelism: Optional[int] = 1000,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
    ray_options_provider: Callable[[int, Any], Dict[str, Any]] = None,
    distributed_dataset_type: Optional[
        DistributedDatasetType
    ] = DistributedDatasetType.RAY_DATASET,
    file_path_column: Optional[str] = None,
    **kwargs,
) -> DistributedDataset:
    """Download all entries in the manifest using the given distributed dataset type.

    Args:
        manifest: The manifest containing the entries to download
        table_type: Dataset type to load the entries into
        max_parallelism: Maximum parallelism to use
        column_names: The list of column names in the table
        include_columns: The list of columns to include in the selection
        file_reader_kwargs_provider: Optional per-content-type provider of file reader kwargs,
            (e.g., to pass in a custom schema for a Parquet file)
        ray_options_provider: Optional provider of Ray options
        distributed_dataset_type: Optional distributed dataset type to use
        file_path_column: Optional file path system column name
        **kwargs: Additional arguments to pass to the file reader

    Returns:
        Distributed dataset
    """
    params = {
        "manifest": manifest,
        "table_type": table_type,
        "max_parallelism": max_parallelism,
        "column_names": column_names,
        "include_columns": include_columns,
        "file_reader_kwargs_provider": file_reader_kwargs_provider,
        "ray_options_provider": ray_options_provider,
        "file_path_column": file_path_column,
        **kwargs,
    }

    if (
        distributed_dataset_type
        and distributed_dataset_type.value == DistributedDatasetType.RAY_DATASET.value
    ):
        result = _download_manifest_entries_ray_data_distributed(**params)
        return result
    elif distributed_dataset_type is not None:
        params["distributed_dataset_type"] = distributed_dataset_type
        return _download_manifest_entries_all_dataset_distributed(**params)
    else:
        raise ValueError(
            f"Distributed dataset type {distributed_dataset_type} not supported."
        )


def _cast_string_view_to_string(table: pa.Table) -> pa.Table:
    """
    Cast any string_view columns to string type for Ray dataset compatibility.

    This addresses compatibility issues where Ray datasets may have trouble with
    string_view columns written by Polars to Feather.

    Args:
        table: PyArrow table that may contain string_view columns

    Returns:
        PyArrow table with string_view columns cast to string type
    """
    if not isinstance(table, pa.Table):
        return table

    schema = table.schema
    has_string_view = False

    # Check if any columns are string_view
    for field in schema:
        if pa.types.is_string_view(field.type):
            has_string_view = True
            break

    if not has_string_view:
        return table

    # Convert to pandas and back to normalize string types
    # This is a workaround since direct casting from string_view to string is not supported
    try:
        pandas_df = table.to_pandas()
        # Convert back to PyArrow table, which should use regular string type
        return pa.Table.from_pandas(pandas_df, preserve_index=False)
    except Exception:
        # If pandas conversion fails, return original table
        return table


def _download_manifest_entries_ray_data_distributed(
    manifest: Manifest,
    table_type: DatasetType = DatasetType.PYARROW,
    max_parallelism: Optional[int] = 1000,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
    ray_options_provider: Callable[[int, Any], Dict[str, Any]] = None,
    file_path_column: Optional[str] = None,
    **kwargs,
) -> DistributedDataset:
    """Download all entries in the manifest into a Ray dataset.

    Args:
        manifest: The manifest containing the entries to download
        table_type: Dataset type to load the entries into
        max_parallelism: Maximum parallelism to use
        column_names: The list of column names in the table
        include_columns: The list of columns to include in the selection
        file_reader_kwargs_provider: Optional per-content-type provider of file reader kwargs,
            (e.g., to pass in a custom schema for a Parquet file)
        ray_options_provider: Optional provider of Ray options
        file_path_column: Optional file path system column name
        **kwargs: Additional arguments to pass to the file reader

    Returns:
        Ray dataset
    """
    table_pending_ids = []
    manifest_entries = manifest.entries

    if manifest_entries:
        table_pending_ids = invoke_parallel(
            manifest_entries,
            download_manifest_entry_ray,
            table_type,
            column_names,
            include_columns,
            file_reader_kwargs_provider,
            max_parallelism=max_parallelism,
            options_provider=ray_options_provider,
            file_path_column=file_path_column,
            **kwargs,  # Pass through kwargs like include_paths
        )

    create_func = _get_table_type_function(
        table_type, TABLE_TYPE_TO_DATASET_CREATE_FUNC_REFS, "dataset create"
    )
    return create_func(table_pending_ids)


def _group_manifest_uris_by_content_type(
    manifest: Manifest, **kwargs
) -> Dict[Tuple[str, str], List[str]]:
    """
    Group manifest URIs by content type and content encoding.

    Args:
        manifest: The manifest containing the entries to group by content type
        **kwargs: Additional arguments to pass to the catalog properties

    Returns:
        Dictionary mapping (content_type, content_encoding) tuples to lists of URIs
    """
    from deltacat.catalog import get_catalog_properties

    catalog_properties = get_catalog_properties(**kwargs)

    uris_by_type = {}

    for entry in manifest.entries or []:
        content_type = entry.meta.content_type
        content_encoding = entry.meta.content_encoding
        key = (content_type, content_encoding)

        if key not in uris_by_type:
            uris_by_type[key] = []

        full_uri = catalog_properties.reconstruct_full_path(entry.uri)
        uris_by_type[key].append(full_uri)

    return uris_by_type


def _download_manifest_entries_all_dataset_distributed(
    manifest: Manifest,
    table_type: DatasetType = DatasetType.PYARROW,
    max_parallelism: Optional[int] = 1000,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
    ray_options_provider: Callable[[int, Any], Dict[str, Any]] = None,
    distributed_dataset_type: Optional[DatasetType] = DatasetType.DAFT,
    file_path_column: Optional[str] = None,
    **kwargs,
) -> DistributedDataset:
    """Download all entries in the manifest into a distributed dataset other than Ray Dataset.

    Args:
        manifest: The manifest containing the entries to download
        table_type: Dataset type to load the entries into
        max_parallelism: Maximum parallelism to use
        column_names: The list of column names in the table
        include_columns: The list of columns to include in the selection
        file_reader_kwargs_provider: Optional per-content-type provider of file reader kwargs,
            (e.g., to pass in a custom schema for a Parquet file)
        ray_options_provider: Optional provider of Ray options
        distributed_dataset_type: Optional distributed dataset type to use
        file_path_column: Optional file path system column name
        **kwargs: Additional arguments to pass to the file reader

    Returns:
        Distributed dataset
    """
    # Group manifest entries by content type instead of validating consistency
    # Filter out table_version_schema from kwargs passed to catalog properties
    filtered_kwargs = _filter_kwargs_for_catalog_properties(kwargs)
    uris_by_content_type = _group_manifest_uris_by_content_type(
        manifest, **filtered_kwargs
    )

    # If only one content type, use the original single-reader logic
    if len(uris_by_content_type) == 1:
        content_type, content_encoding = next(iter(uris_by_content_type.keys()))
        uris = next(iter(uris_by_content_type.values()))

        # Keep table_version_schema for the reader, but filter other system kwargs
        reader_kwargs = _filter_kwargs_for_reader_functions(kwargs)

        try:
            reader_func = DISTRIBUTED_DATASET_TYPE_TO_READER_FUNC[
                distributed_dataset_type.value
            ]
        except KeyError:
            raise ValueError(
                f"Unsupported distributed dataset type={distributed_dataset_type}. "
                f"Supported types: {list(DISTRIBUTED_DATASET_TYPE_TO_READER_FUNC.keys())}"
            )

        return reader_func(
            uris=uris,
            content_type=content_type,
            content_encoding=content_encoding,
            column_names=column_names,
            include_columns=include_columns,
            read_func_kwargs_provider=file_reader_kwargs_provider,
            ray_options_provider=ray_options_provider,
            file_path_column=file_path_column,
            **reader_kwargs,
        )

    # Multiple content types - read each group and union them (only for Daft)
    if distributed_dataset_type != DistributedDatasetType.DAFT:
        raise ValueError(
            f"Mixed content types are only supported for Daft datasets. "
            f"Got {len(uris_by_content_type)} different content types with dataset type {distributed_dataset_type}"
        )

    # Keep table_version_schema for the reader, but filter other system kwargs
    reader_kwargs = _filter_kwargs_for_reader_functions(kwargs)

    try:
        reader_func = DISTRIBUTED_DATASET_TYPE_TO_READER_FUNC[
            distributed_dataset_type.value
        ]
    except KeyError:
        raise ValueError(
            f"Unsupported distributed dataset type={distributed_dataset_type}. "
            f"Supported types: {list(DISTRIBUTED_DATASET_TYPE_TO_READER_FUNC.keys())}"
        )

    # Read each content type group into a separate DataFrame
    dataframes = []
    for (content_type, content_encoding), uris in uris_by_content_type.items():
        df = reader_func(
            uris=uris,
            content_type=content_type,
            content_encoding=content_encoding,
            column_names=column_names,
            include_columns=include_columns,
            read_func_kwargs_provider=file_reader_kwargs_provider,
            ray_options_provider=ray_options_provider,
            file_path_column=file_path_column,
            **reader_kwargs,
        )
        dataframes.append(df)

    # Union all DataFrames using Daft's union_all
    if len(dataframes) == 1:
        return dataframes[0]

    result = dataframes[0]
    for df in dataframes[1:]:
        result = result.union_all(df)

    return result


def _download_manifest_entries_parallel(
    manifest: Manifest,
    table_type: DatasetType = DatasetType.PYARROW,
    max_parallelism: Optional[int] = None,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
    file_path_column: Optional[str] = None,
    **kwargs,
) -> LocalDataset:
    """Download all entries in the manifest into a local dataset using multiprocessing.

    Args:
        manifest: The manifest containing the entries to download
        table_type: Dataset type to load the entries into
        max_parallelism: Maximum parallel processes to use for entry downloads
        column_names: The list of column names in the table
        include_columns: The list of columns to include in the selection
        file_reader_kwargs_provider: Optional per-content-type provider of file reader kwargs,
            (e.g., to pass in a custom schema for a Parquet file)
        file_path_column: Optional file path system column name
        **kwargs: Additional arguments to pass to the file reader

    Returns:
        Local dataset
    """
    download_args = _prepare_download_arguments(
        table_type,
        column_names,
        include_columns,
        file_reader_kwargs_provider,
        file_path_column,
        **kwargs,
    )

    entries_to_process = []
    for e in manifest.entries:
        manifest_entry = _reconstruct_manifest_entry_uri(e, **kwargs)
        entries_to_process.append(manifest_entry)

    tables = []
    pool = multiprocessing.Pool(max_parallelism)

    downloader = partial(download_manifest_entry, **download_args)
    for table in pool.map(downloader, entries_to_process):
        tables.append(table)
    return tables


def download_manifest_entry(
    manifest_entry: ManifestEntry,
    table_type: DatasetType = DatasetType.PYARROW,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
    content_type: Optional[ContentType] = None,
    content_encoding: Optional[ContentEncoding] = None,
    filesystem: Optional[pyarrow.fs.FileSystem] = None,
    file_path_column: Optional[str] = None,
    **kwargs,
) -> LocalTable:
    """Download a single entry in the manifest into a local table.

    Args:
        manifest_entry: The manifest entry to download
        table_type: Dataset type to load the entry into
        column_names: The list of column names in the table
        include_columns: The list of columns to include in the selection
        file_reader_kwargs_provider: Optional per-content-type provider of file reader kwargs,
            (e.g., to pass in a custom schema for a Parquet file)
        content_type: Optional content type of the file
        content_encoding: Optional content encoding of the file
        filesystem: Optional PyArrow filesystem to use to read the file
        file_path_column: Optional file path system column name
        **kwargs: Additional arguments to pass to the file reader

    Returns:
        Local table
    """
    # Extract manifest metadata
    (
        extracted_content_type,
        extracted_content_encoding,
        path,
    ) = _extract_content_metadata(manifest_entry)
    content_type = content_type or extracted_content_type
    content_encoding = content_encoding or extracted_content_encoding

    # Extract partial download parameters
    partial_file_download_params = _extract_partial_download_params(manifest_entry)

    # Filter kwargs and process file path column
    reader_kwargs = _filter_kwargs_for_external_readers(kwargs)
    processed_include_columns = _remove_file_path_column(
        include_columns, file_path_column
    )

    # Create retry wrapper and read file
    retrying = _create_retry_wrapper()
    table = retrying(
        read_file,
        path,
        content_type,
        content_encoding,
        table_type,
        column_names,
        processed_include_columns,
        file_reader_kwargs_provider,
        partial_file_download_params,
        filesystem,
        **reader_kwargs,
    )

    # Add file path column if requested
    if file_path_column:
        if isinstance(table, papq.ParquetFile):
            logger.warning(
                f"Skipping file_path_column '{file_path_column}' for lazily materialized ParquetFile. "
                f"File path information can be retrieved from the ParquetFile object's metadata. "
                f"Use read_as=DatasetType.PYARROW to materialize with file path column."
            )
        else:
            table = append_column_to_table(table, file_path_column, manifest_entry.uri)

    return table


@categorize_errors
def read_file(
    path: str,
    content_type: ContentType,
    content_encoding: ContentEncoding = ContentEncoding.IDENTITY,
    table_type: DatasetType = DatasetType.PYARROW,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
    partial_file_download_params: Optional[PartialFileDownloadParams] = None,
    filesystem: Optional[pyarrow.fs.FileSystem] = None,
    **kwargs,
) -> LocalTable:
    """Read a file into a local table.

    Args:
        path: The path to the file to read
        content_type: The content type of the file
        content_encoding: The content encoding of the file
        table_type: Dataset type to load the file into
        column_names: The list of column names in the table
        include_columns: The list of columns to include in the selection
        file_reader_kwargs_provider: Optional per-content-type provider of file reader kwargs,
            (e.g., to pass in a custom schema for a Parquet file)
        partial_file_download_params: Optional partial file download parameters
        filesystem: Optional PyArrow filesystem to use to read the file
        **kwargs: Additional arguments to pass to the file reader

    Returns:
        Local table
    """
    reader = TABLE_TYPE_TO_READER_FUNC[table_type.value]
    try:
        table = reader(
            path,
            content_type.value,
            content_encoding.value,
            filesystem,
            column_names,
            include_columns,
            file_reader_kwargs_provider,
            partial_file_download_params,
            **kwargs,
        )
        return table
    except RETRYABLE_TRANSIENT_ERRORS as e:
        _handle_retryable_error(e, path, "download", RetryableDownloadTableError)
    except BaseException as e:
        _handle_non_retryable_error(
            e,
            path,
            "read",
            NonRetryableDownloadTableError,
            f"and content_type={content_type} and encoding={content_encoding}",
        )



import pytest
import tempfile
import shutil
import uuid
from unittest import mock
import os

from deltacat.catalog import (
    CatalogProperties,
    Catalog,
    clear_catalogs,
    get_catalog,
    init,
    init_local,
    is_initialized,
    put_catalog,
)
from deltacat.experimental.catalog.iceberg import impl as IcebergCatalog
from pyiceberg.catalog import Catalog as PyIcebergCatalog

from deltacat.experimental.catalog.iceberg import IcebergCatalogConfig

from pyiceberg.catalog import CatalogType


# Test module to mock a catalog implementation
class MockCatalogImpl:
    @staticmethod
    def initialize(config, *args, **kwargs):
        # Return some state that the catalog would normally maintain
        return {
            "initialized": True,
            "config": config,
            "args": args,
            "kwargs": kwargs,
        }


@pytest.fixture(scope="function")
def reset_catalogs():
    clear_catalogs()


class TestCatalog:
    """Tests for the Catalog class itself, without Ray initialization."""

    def test_catalog_constructor(self):
        """Test that the Catalog constructor correctly initializes with the given implementation."""
        catalog = Catalog(impl=MockCatalogImpl)

        assert catalog.impl == MockCatalogImpl

        # Check that inner state was correctly initialized
        # This just asserts that kwargs were plumbed through from Catalog constructor
        assert catalog.inner["initialized"]
        assert catalog.inner["config"] is None
        assert catalog.inner["args"] == ()
        assert catalog.inner["kwargs"] == {}

    def test_iceberg_factory_method(self):
        """Test the iceberg factory method correctly creates an Iceberg catalog."""
        # Create a mock for the Iceberg catalog module
        with mock.patch(
            "deltacat.experimental.catalog.iceberg.impl.IcebergCatalog"
        ) as mock_iceberg_catalog:
            # Configure the mock to return a known value when initialize is called
            mock_iceberg_catalog.initialize.return_value = {"iceberg": True}

            # Create an Iceberg catalog config and invoke iceberg factory method
            config = IcebergCatalogConfig(type=CatalogType.IN_MEMORY, properties={})
            catalog = IcebergCatalog.from_config(config)

            # Check that the implementation is set to iceberg_catalog
            assert catalog.impl == mock_iceberg_catalog
            # Check that the inner state is set to the output of initialize
            assert catalog.inner == {"iceberg": True}


class TestCatalogsIntegration:
    """Integration tests for Default catalog functionality."""

    temp_dir = None

    @classmethod
    def setup_class(cls):
        cls.temp_dir = tempfile.mkdtemp()
        # Other tests are going to have initialized ray catalog. Initialize here to ensure
        # that when this test class is run individuall it mimicks running with other tests
        catalog = Catalog(impl=MockCatalogImpl)
        init(
            catalog,
            force=True,
        )

    @classmethod
    def teardown_class(cls):
        if cls.temp_dir and os.path.exists(cls.temp_dir):
            shutil.rmtree(cls.temp_dir)

    def test_init_single_catalog(self, reset_catalogs):
        """Test initializing a single catalog."""

        catalog = Catalog(impl=MockCatalogImpl)

        # Initialize with a single catalog and Ray init args including the namespace
        init(catalog, force=True)

        assert is_initialized()

        # Get the default catalog and check it's the same one we initialized with
        retrieved_catalog = get_catalog()
        assert retrieved_catalog.impl == MockCatalogImpl
        assert retrieved_catalog.inner["initialized"]

    def test_init_multiple_catalogs(self, reset_catalogs):
        """Test initializing multiple catalogs."""
        # Create catalogs
        catalog1 = Catalog(impl=MockCatalogImpl, id=1)
        catalog2 = Catalog(impl=MockCatalogImpl, id=2)

        # Initialize with multiple catalogs and Ray init args including the namespace
        catalogs_dict = {"catalog1": catalog1, "catalog2": catalog2}
        init(catalogs_dict, force=True)

        assert is_initialized()

        # Get catalogs by name and check they're the same ones we initialized with
        retrieved_catalog1 = get_catalog("catalog1")
        assert retrieved_catalog1.impl == MockCatalogImpl
        assert retrieved_catalog1.inner["kwargs"]["id"] == 1

        retrieved_catalog2 = get_catalog("catalog2")
        assert retrieved_catalog2.impl == MockCatalogImpl
        assert retrieved_catalog2.inner["kwargs"]["id"] == 2

    def test_init_with_default_catalog_name(self, reset_catalogs):
        """Test initializing with a specified default catalog name."""
        # Create catalogs
        catalog1 = Catalog(impl=MockCatalogImpl, id=1)
        catalog2 = Catalog(impl=MockCatalogImpl, id=2)

        # Initialize with multiple catalogs and specify a default
        catalogs_dict = {"catalog1": catalog1, "catalog2": catalog2}
        init(
            catalogs_dict,
            default="catalog2",
            force=True,
        )

        # Get the default catalog and check it's catalog2
        default_catalog = get_catalog()
        assert default_catalog.impl == MockCatalogImpl
        assert default_catalog.inner["kwargs"]["id"] == 2

    def test_put_catalog(self, reset_catalogs):
        """Test adding a catalog after initialization."""
        # Initialize with a single catalog
        catalog1 = Catalog(impl=MockCatalogImpl, id=1)
        catalog2 = Catalog(impl=MockCatalogImpl, id=2)
        init({"catalog1": catalog1}, force=True)

        # Add a second catalog
        put_catalog("catalog2", catalog2)

        # Check both catalogs are available
        retrieved_catalog1 = get_catalog("catalog1")
        assert retrieved_catalog1.inner["kwargs"]["id"] == 1

        retrieved_catalog2 = get_catalog("catalog2")
        assert retrieved_catalog2.inner["kwargs"]["id"] == 2

    def test_put_catalog_that_already_exists(self, reset_catalogs):
        catalog = Catalog(impl=MockCatalogImpl, id=1)
        catalog2 = Catalog(impl=MockCatalogImpl, id=2)
        put_catalog(
            "test_catalog",
            catalog,
            id=1,
        )

        # Try to add another catalog with the same name. Should not error
        put_catalog(
            "test_catalog",
            catalog2,
        )

        retrieved_catalog = get_catalog("test_catalog")
        assert retrieved_catalog.inner["kwargs"]["id"] == 2

        # If fail_if_exists, put call should fail
        with pytest.raises(ValueError):
            put_catalog(
                "test_catalog",
                catalog,
                fail_if_exists=True,
            )

    def test_get_catalog_nonexistent(self, reset_catalogs):
        """Test that trying to get a nonexistent catalog raises an error."""
        # Initialize with a catalog
        catalog = Catalog(impl=MockCatalogImpl)
        init({"test_catalog": catalog}, force=True)

        # Try to get a nonexistent catalog
        with pytest.raises(ValueError):
            get_catalog("nonexistent")

    def test_get_catalog_no_default(self, reset_catalogs):
        """Test that trying to get the default catalog when none is set raises an error."""
        # Initialize with multiple catalogs but no default
        catalog1 = Catalog(impl=MockCatalogImpl, id=1)
        catalog2 = Catalog(impl=MockCatalogImpl, id=2)
        init({"catalog1": catalog1, "catalog2": catalog2}, force=True)

        # Try to get the default catalog
        with pytest.raises(ValueError):
            get_catalog()

    def test_init_local(self, reset_catalogs):
        """Test that init_local() creates a default local catalog."""
        # Initialize with default local catalog
        init_local(force=True)

        assert is_initialized()

        # Should be able to get the default catalog
        default_catalog = get_catalog()
        assert default_catalog is not None

        # The default catalog should be accessible by name "default"
        named_catalog = get_catalog("default")
        assert named_catalog is not None
        assert named_catalog.impl.__name__ == "deltacat.catalog.main.impl"

    def test_init_local_with_path(self, reset_catalogs):
        """Test that init_local(path) creates a default local catalog with specified path."""
        # Create a temporary directory for the test
        custom_path = tempfile.mkdtemp()

        try:
            # Initialize with custom path
            init_local(path=custom_path, force=True)

            assert is_initialized()

            # Should be able to get the default catalog
            default_catalog = get_catalog()
            assert default_catalog is not None

            # The default catalog should be accessible by name "default"
            named_catalog = get_catalog("default")
            assert named_catalog is not None
            assert named_catalog.impl.__name__ == "deltacat.catalog.main.impl"

            # Verify the catalog is using the custom path
            catalog_properties = named_catalog.inner
            assert catalog_properties.root == custom_path

        finally:
            # Clean up the temporary directory
            if os.path.exists(custom_path):
                shutil.rmtree(custom_path)

    def test_default_catalog_initialization(self, reset_catalogs):
        """Test that a Default catalog can be initialized and accessed using the factory method."""
        from deltacat.catalog.model.properties import CatalogProperties

        catalog_name = str(uuid.uuid4())

        # Create the catalog properties
        config = CatalogProperties(root=self.temp_dir)

        # Create the catalog
        catalog = Catalog(config)

        # Initialize DeltaCAT with this catalog
        init({catalog_name: catalog}, force=True)

        # Retrieve the catalog and verify it's the same one
        retrieved_catalog = get_catalog(catalog_name)
        assert retrieved_catalog.impl.__name__ == "deltacat.catalog.main.impl"
        assert isinstance(retrieved_catalog.inner, CatalogProperties)
        assert retrieved_catalog.inner.root == self.temp_dir

    def test_default_catalog_initialization_from_kwargs(self, reset_catalogs):

        catalog_name = str(uuid.uuid4())

        # Initialize DeltaCAT with this catalog
        put_catalog(
            catalog_name,
            Catalog(root="test_root"),
        )

        # Retrieve the catalog and verify it's the same one
        retrieved_catalog = get_catalog(catalog_name)
        assert retrieved_catalog.impl.__name__ == "deltacat.catalog.main.impl"
        assert isinstance(retrieved_catalog.inner, CatalogProperties)
        assert retrieved_catalog.inner.root == "test_root"

    def test_iceberg_catalog_initialization(self, reset_catalogs):
        """Test that an Iceberg catalog can be initialized and accessed."""
        catalog_name = str(uuid.uuid4())

        # Create the Iceberg catalog config
        config = IcebergCatalogConfig(
            type=CatalogType.IN_MEMORY, properties={"warehouse": self.temp_dir}
        )

        # Create the catalog using the factory method
        catalog = IcebergCatalog.from_config(config)

        put_catalog(catalog_name, catalog)

        # Retrieve the catalog and verify it's the same one
        retrieved_catalog = get_catalog(catalog_name)
        assert (
            retrieved_catalog.impl.__name__
            == "deltacat.experimental.catalog.iceberg.impl"
        )
        assert isinstance(retrieved_catalog.inner, PyIcebergCatalog)



import numpy as np
import functools
import logging
import ray
import time
import json
from math import ceil
from urllib.parse import urlparse
import pyarrow

from deltacat.compute.compactor import (
    PyArrowWriteResult,
    HighWatermark,
    RoundCompletionInfo,
)
from deltacat.utils.filesystem import resolve_path_and_filesystem
from deltacat.compute.compactor.model.compactor_version import CompactorVersion
from deltacat.compute.compactor_v2.model.evaluate_compaction_result import (
    ExecutionCompactionResult,
)
from deltacat.compute.compactor_v2.model.merge_file_group import (
    RemoteMergeFileGroupsProvider,
)
from deltacat.compute.compactor_v2.model.hash_bucket_input import HashBucketInput

from deltacat import logs
from deltacat.compute.compactor_v2.model.merge_input import MergeInput
from deltacat.compute.compactor_v2.model.merge_result import MergeResult
from deltacat.compute.compactor_v2.model.hash_bucket_result import HashBucketResult
from deltacat.compute.compactor.model.materialize_result import MaterializeResult
from deltacat.compute.compactor_v2.utils.merge import (
    generate_local_merge_input,
)
from deltacat.compute.compactor_v2.utils.task_options import (
    hash_bucket_resource_options_provider,
)
from deltacat.compute.compactor.utils import round_completion_reader as rci
from deltacat.compute.compactor import DeltaAnnotated
from deltacat.compute.compactor_v2.utils.delta import contains_delete_deltas
from deltacat.compute.compactor_v2.deletes.delete_strategy import (
    DeleteStrategy,
)
from deltacat.compute.compactor_v2.deletes.delete_file_envelope import (
    DeleteFileEnvelope,
)
from deltacat.compute.compactor_v2.deletes.utils import prepare_deletes

from deltacat.storage import (
    Delta,
    DeltaType,
    DeltaLocator,
    Partition,
    PartitionLocator,
    Manifest,
    Stream,
    StreamLocator,
)
from deltacat.compute.compactor.model.compact_partition_params import (
    CompactPartitionParams,
)
from deltacat.utils.ray_utils.concurrency import (
    invoke_parallel,
    task_resource_options_provider,
)
from deltacat.compute.compactor_v2.steps import merge as mg
from deltacat.compute.compactor_v2.steps import hash_bucket as hb
from deltacat.compute.compactor_v2.utils import io

from typing import List, Optional, Union
from collections import defaultdict
from deltacat.compute.compactor.model.compaction_session_audit_info import (
    CompactionSessionAuditInfo,
)
from deltacat.compute.compactor_v2.utils.task_options import (
    merge_resource_options_provider,
    local_merge_resource_options_provider,
)


logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def _get_rci_source_partition_locator(
    params: CompactPartitionParams,
) -> PartitionLocator:
    return params.rebase_source_partition_locator or params.source_partition_locator


def _is_inplace_compacted(
    rci_source_partition_locator: PartitionLocator,
    destination_partition_locator: PartitionLocator,
) -> bool:
    return (
        rci_source_partition_locator.partition_values
        == destination_partition_locator.partition_values
        and rci_source_partition_locator.stream_id
        == destination_partition_locator.stream_id
    )


def _fetch_compaction_metadata(
    params: CompactPartitionParams,
) -> tuple[Optional[Manifest], Optional[RoundCompletionInfo]]:

    # read the results from any previously completed compaction round
    round_completion_info: Optional[RoundCompletionInfo] = None
    high_watermark: Optional[Union[HighWatermark, int]] = None
    previous_compacted_delta_manifest: Optional[Manifest] = None

    if not params.rebase_source_partition_locator:
        round_completion_info = rci.read_round_completion_info(
            source_partition_locator=params.source_partition_locator,
            destination_partition_locator=params.destination_partition_locator,
            deltacat_storage=params.deltacat_storage,
            deltacat_storage_kwargs=params.deltacat_storage_kwargs,
        )
        if not round_completion_info:
            logger.info(
                "Both rebase partition and round completion file not found. Performing an entire backfill on source."
            )
        else:
            compacted_delta_locator = round_completion_info.compacted_delta_locator

            previous_compacted_delta_manifest = (
                params.deltacat_storage.get_delta_manifest(
                    compacted_delta_locator, **params.deltacat_storage_kwargs
                )
            )

            high_watermark = round_completion_info.high_watermark
            logger.info(f"Setting round completion high watermark: {high_watermark}")
            assert (
                params.hash_bucket_count == round_completion_info.hash_bucket_count
            ), (
                "Partition hash bucket count for compaction has changed. "
                "Rebase compaction with the desired hash bucket count before running another incremental compaction. "
                f"Hash bucket count in RCI={round_completion_info.hash_bucket_count} "
                f"!= hash bucket count in params={params.hash_bucket_count}."
            )

        logger.info(f"Round completion file: {round_completion_info}")
    return (
        previous_compacted_delta_manifest,
        round_completion_info,
    )


def _build_uniform_deltas(
    params: CompactPartitionParams,
    mutable_compaction_audit: CompactionSessionAuditInfo,
    input_deltas: List[Delta],
    delta_discovery_start: float,
) -> tuple[List[DeltaAnnotated], DeleteStrategy, List[DeleteFileEnvelope]]:

    delete_strategy: Optional[DeleteStrategy] = None
    delete_file_envelopes: Optional[List[DeleteFileEnvelope]] = None
    delete_file_size_bytes: int = 0
    if contains_delete_deltas(input_deltas):
        input_deltas, delete_file_envelopes, delete_strategy = prepare_deletes(
            params, input_deltas
        )
        for delete_file_envelope in delete_file_envelopes:
            delete_file_size_bytes += delete_file_envelope.table_size_bytes
        logger.info(
            f" Input deltas contain {DeltaType.DELETE}-type deltas. Total delete file size={delete_file_size_bytes}."
            f" Total length of delete file envelopes={len(delete_file_envelopes)}"
        )
    uniform_deltas: List[DeltaAnnotated] = io.create_uniform_input_deltas(
        input_deltas=input_deltas,
        hash_bucket_count=params.hash_bucket_count,
        compaction_audit=mutable_compaction_audit,
        compact_partition_params=params,
        all_column_names=params.all_column_names,
        deltacat_storage=params.deltacat_storage,
        deltacat_storage_kwargs=params.deltacat_storage_kwargs,
    )
    delta_discovery_end: float = time.monotonic()

    mutable_compaction_audit.set_uniform_deltas_created(len(uniform_deltas))
    mutable_compaction_audit.set_delta_discovery_time_in_seconds(
        delta_discovery_end - delta_discovery_start
    )

    _upload_audit_data(
        params,
        mutable_compaction_audit,
    )

    return (
        uniform_deltas,
        delete_strategy,
        delete_file_envelopes,
    )


def _group_uniform_deltas(
    params: CompactPartitionParams, uniform_deltas: List[DeltaAnnotated]
) -> List[List[DeltaAnnotated]]:
    num_deltas = len(uniform_deltas)
    num_rounds = params.num_rounds
    if num_rounds == 1:
        return [uniform_deltas]
    assert (
        num_rounds > 0
    ), f"num_rounds parameter should be greater than zero but is {params.num_rounds}"
    assert (
        num_rounds <= num_deltas
    ), f"{params.num_rounds} rounds should be less than the number of uniform deltas, which is {len(uniform_deltas)}"
    size = ceil(num_deltas / num_rounds)
    uniform_deltas_grouped = list(
        map(
            lambda x: uniform_deltas[x * size : x * size + size],
            list(range(num_rounds)),
        )
    )
    num_deltas_after_grouping = sum(len(sublist) for sublist in uniform_deltas_grouped)
    assert (
        num_deltas_after_grouping == num_deltas
    ), f"uniform_deltas_grouped expected to have {num_deltas} deltas, but has {num_deltas_after_grouping}"
    return uniform_deltas_grouped


def _stage_new_partition(params: CompactPartitionParams) -> Partition:
    compacted_stream_locator: Optional[
        StreamLocator
    ] = params.destination_partition_locator.stream_locator
    compacted_stream: Stream = params.deltacat_storage.get_stream(
        compacted_stream_locator.namespace,
        compacted_stream_locator.table_name,
        compacted_stream_locator.table_version,
        **params.deltacat_storage_kwargs,
    )
    compacted_partition: Partition = params.deltacat_storage.stage_partition(
        compacted_stream,
        params.destination_partition_locator.partition_values,
        **params.deltacat_storage_kwargs,
    )
    return compacted_partition


def _run_hash_and_merge(
    params: CompactPartitionParams,
    uniform_deltas: List[DeltaAnnotated],
    round_completion_info: RoundCompletionInfo,
    delete_strategy: Optional[DeleteStrategy],
    delete_file_envelopes: Optional[List[DeleteFileEnvelope]],
    mutable_compaction_audit: CompactionSessionAuditInfo,
    previous_compacted_delta_manifest: Optional[Manifest],
    compacted_partition: Partition,
) -> List[MergeResult]:
    telemetry_time_hb = 0
    total_input_records_count = np.int64(0)
    total_hb_record_count = np.int64(0)
    if params.hash_bucket_count == 1:
        logger.info("Hash bucket count set to 1. Running local merge")
        merge_start: float = time.monotonic()
        merge_results, total_input_records_count = _run_local_merge(
            params,
            uniform_deltas,
            compacted_partition,
            round_completion_info,
            delete_strategy,
            delete_file_envelopes,
            mutable_compaction_audit,
            previous_compacted_delta_manifest,
            total_input_records_count,
        )
        merge_invoke_end = time.monotonic()
    else:
        # hash bucket
        hb_start = time.monotonic()
        all_hash_group_idx_to_obj_id = defaultdict(list)
        all_hash_group_idx_to_size_bytes = defaultdict(int)
        all_hash_group_idx_to_num_rows = defaultdict(int)
        (hb_results, hb_invoke_end) = _hash_bucket(params, uniform_deltas)
        hb_end = time.monotonic()

        # we use time.time() here because time.monotonic() has no reference point
        # whereas time.time() measures epoch seconds. Hence, it will be reasonable
        # to compare time.time()s captured in different nodes.
        hb_results_retrieved_at = time.time()

        telemetry_time_hb = mutable_compaction_audit.save_step_stats(
            CompactionSessionAuditInfo.HASH_BUCKET_STEP_NAME,
            hb_results,
            hb_results_retrieved_at,
            hb_invoke_end - hb_start,
            hb_end - hb_start,
        )

        _upload_audit_data(
            params,
            mutable_compaction_audit,
        )

        hb_data_processed_size_bytes = np.int64(0)

        # initialize all hash groups
        for hb_group in range(params.hash_group_count):
            all_hash_group_idx_to_num_rows[hb_group] = 0
            all_hash_group_idx_to_obj_id[hb_group] = []
            all_hash_group_idx_to_size_bytes[hb_group] = 0

        for hb_result in hb_results:
            hb_data_processed_size_bytes += hb_result.hb_size_bytes
            total_input_records_count += hb_result.hb_record_count
            for hash_group_index, object_id_size_tuple in enumerate(
                hb_result.hash_bucket_group_to_obj_id_tuple
            ):
                if object_id_size_tuple:
                    all_hash_group_idx_to_obj_id[hash_group_index].append(
                        object_id_size_tuple[0],
                    )
                    all_hash_group_idx_to_size_bytes[
                        hash_group_index
                    ] += object_id_size_tuple[1].item()
                    all_hash_group_idx_to_num_rows[
                        hash_group_index
                    ] += object_id_size_tuple[2].item()
        logger.info(
            f"Got {total_input_records_count} hash bucket records from hash bucketing step..."
        )

        total_hb_record_count = total_input_records_count
        mutable_compaction_audit.set_hash_bucket_processed_size_bytes(
            hb_data_processed_size_bytes.item()
        )

        # BSP Step 2: Merge
        # NOTE: DELETE-type deltas are stored in Plasma object store
        # in prepare_deletes and therefore don't need to included
        # in merge task resource estimation
        merge_start = time.monotonic()
        merge_results, merge_invoke_end = _merge(
            params,
            task_resource_options_provider,
            merge_resource_options_provider,
            all_hash_group_idx_to_size_bytes,
            all_hash_group_idx_to_num_rows,
            round_completion_info,
            previous_compacted_delta_manifest,
            all_hash_group_idx_to_obj_id,
            compacted_partition,
            delete_strategy,
            delete_file_envelopes,
        )
    logger.info(f"Got {len(merge_results)} merge results.")

    merge_results_retrieved_at: float = time.time()
    merge_end: float = time.monotonic()

    total_dd_record_count = sum([ddr.deduped_record_count for ddr in merge_results])
    total_deleted_record_count = sum(
        [ddr.deleted_record_count for ddr in merge_results]
    )
    logger.info(
        f"Deduped {total_dd_record_count} records and deleted {total_deleted_record_count} records..."
    )

    mutable_compaction_audit.set_input_records(total_input_records_count.item())

    telemetry_time_merge = mutable_compaction_audit.save_step_stats(
        CompactionSessionAuditInfo.MERGE_STEP_NAME,
        merge_results,
        merge_results_retrieved_at,
        merge_invoke_end - merge_start,
        merge_end - merge_start,
    )

    mutable_compaction_audit.set_records_deduped(total_dd_record_count.item())
    mutable_compaction_audit.set_records_deleted(total_deleted_record_count.item())
    record_info_msg: str = (
        f"Hash bucket records: {total_hb_record_count},"
        f" Deduped records: {total_dd_record_count}, "
        f" Deleted records: {total_deleted_record_count}, "
    )
    logger.info(record_info_msg)
    telemetry_this_round = telemetry_time_hb + telemetry_time_merge
    previous_telemetry = (
        mutable_compaction_audit.telemetry_time_in_seconds
        if mutable_compaction_audit.telemetry_time_in_seconds
        else 0.0
    )

    mutable_compaction_audit.set_telemetry_time_in_seconds(
        telemetry_this_round + previous_telemetry
    )
    if params.num_rounds > 1:
        logger.info(
            f"Detected number of rounds to be {params.num_rounds}, preparing to clear object store..."
        )
        params.object_store.clear()
    else:
        logger.info(
            f"Detected number of rounds to be {params.num_rounds}, not cleaning up object store..."
        )

    return merge_results


def _merge(
    params: CompactPartitionParams,
    task_resource_options_provider: callable,
    merge_resource_options_provider: callable,
    all_hash_group_idx_to_size_bytes: dict,
    all_hash_group_idx_to_num_rows: dict,
    round_completion_info: RoundCompletionInfo,
    previous_compacted_delta_manifest: Manifest,
    all_hash_group_idx_to_obj_id: dict,
    compacted_partition: Partition,
    delete_strategy: DeleteStrategy,
    delete_file_envelopes: List[DeleteFileEnvelope],
) -> tuple[List[MergeResult], float]:
    merge_options_provider = functools.partial(
        task_resource_options_provider,
        pg_config=params.pg_config,
        resource_amount_provider=merge_resource_options_provider,
        num_hash_groups=params.hash_group_count,
        hash_group_size_bytes=all_hash_group_idx_to_size_bytes,
        hash_group_num_rows=all_hash_group_idx_to_num_rows,
        total_memory_buffer_percentage=params.total_memory_buffer_percentage,
        round_completion_info=round_completion_info,
        compacted_delta_manifest=previous_compacted_delta_manifest,
        primary_keys=params.primary_keys,
        ray_custom_resources=params.ray_custom_resources,
        memory_logs_enabled=params.memory_logs_enabled,
        estimate_resources_params=params.estimate_resources_params,
    )

    # set previous compacted delta manifest on input so that we don't need a transaction to retrieve it
    if round_completion_info:
        previous_compacted_delta_manifest = params.deltacat_storage.get_delta_manifest(
            round_completion_info.compacted_delta_locator,
            **params.deltacat_storage_kwargs,
        )

    # create a copy of deltacat storage kwargs without any parent transaction context
    # (can't be serialized by Ray, and we're only downloading already-resolved manifest entries)
    deltacat_storage_kwargs_copy = {
        k: v for k, v in params.deltacat_storage_kwargs.items() if k != "transaction"
    }

    def merge_input_provider(index, item) -> dict[str, MergeInput]:
        return {
            "input": MergeInput.of(
                merge_file_groups_provider=RemoteMergeFileGroupsProvider(
                    hash_group_index=item[0],
                    dfe_groups_refs=item[1],
                    hash_bucket_count=params.hash_bucket_count,
                    num_hash_groups=params.hash_group_count,
                    object_store=params.object_store,
                ),
                write_to_partition=compacted_partition,
                compacted_file_content_type=params.compacted_file_content_type,
                primary_keys=params.primary_keys,
                all_column_names=params.all_column_names,
                sort_keys=params.sort_keys,
                merge_task_index=index,
                drop_duplicates=params.drop_duplicates,
                max_records_per_output_file=params.records_per_compacted_file,
                enable_profiler=params.enable_profiler,
                metrics_config=params.metrics_config,
                table_writer_kwargs=params.table_writer_kwargs,
                read_kwargs_provider=params.read_kwargs_provider,
                round_completion_info=round_completion_info,
                object_store=params.object_store,
                deltacat_storage=params.deltacat_storage,
                deltacat_storage_kwargs=deltacat_storage_kwargs_copy,
                delete_strategy=delete_strategy,
                delete_file_envelopes=delete_file_envelopes,
                memory_logs_enabled=params.memory_logs_enabled,
                disable_copy_by_reference=params.disable_copy_by_reference,
                hash_bucket_count=params.hash_bucket_count,
                original_fields=params.original_fields,
                compacted_manifest=previous_compacted_delta_manifest,
            )
        }

    merge_tasks_pending = invoke_parallel(
        items=all_hash_group_idx_to_obj_id.items(),
        ray_task=mg.merge,
        max_parallelism=params.task_max_parallelism,
        options_provider=merge_options_provider,
        kwargs_provider=merge_input_provider,
    )
    merge_invoke_end = time.monotonic()
    logger.info(f"Getting {len(merge_tasks_pending)} merge results...")
    merge_results: List[MergeResult] = ray.get(merge_tasks_pending)

    return merge_results, merge_invoke_end


def _hash_bucket(
    params: CompactPartitionParams,
    uniform_deltas: List[DeltaAnnotated],
) -> tuple[List[HashBucketResult], float]:

    hb_options_provider = functools.partial(
        task_resource_options_provider,
        pg_config=params.pg_config,
        resource_amount_provider=hash_bucket_resource_options_provider,
        previous_inflation=params.previous_inflation,
        average_record_size_bytes=params.average_record_size_bytes,
        total_memory_buffer_percentage=params.total_memory_buffer_percentage,
        primary_keys=params.primary_keys,
        ray_custom_resources=params.ray_custom_resources,
        memory_logs_enabled=params.memory_logs_enabled,
        estimate_resources_params=params.estimate_resources_params,
    )

    # create a copy of deltacat storage kwargs without any parent transaction context
    # (can't be serialized by Ray, and we're only downloading already-resolved manifest entries)
    deltacat_storage_kwargs_copy = {
        k: v for k, v in params.deltacat_storage_kwargs.items() if k != "transaction"
    }

    def hash_bucket_input_provider(index, item) -> dict[str, HashBucketInput]:
        return {
            "input": HashBucketInput.of(
                item,
                primary_keys=params.primary_keys,
                hb_task_index=index,
                num_hash_buckets=params.hash_bucket_count,
                num_hash_groups=params.hash_group_count,
                all_column_names=params.all_column_names,
                enable_profiler=params.enable_profiler,
                metrics_config=params.metrics_config,
                read_kwargs_provider=params.read_kwargs_provider,
                object_store=params.object_store,
                deltacat_storage=params.deltacat_storage,
                deltacat_storage_kwargs=deltacat_storage_kwargs_copy,
                memory_logs_enabled=params.memory_logs_enabled,
            )
        }

    hb_tasks_pending = invoke_parallel(
        items=uniform_deltas,
        ray_task=hb.hash_bucket,
        max_parallelism=params.task_max_parallelism,
        options_provider=hb_options_provider,
        kwargs_provider=hash_bucket_input_provider,
    )
    hb_invoke_end = time.monotonic()

    logger.info(f"Getting {len(hb_tasks_pending)} hash bucket results...")
    hb_results: List[HashBucketResult] = ray.get(hb_tasks_pending)
    logger.info(f"Got {len(hb_results)} hash bucket results.")

    return (hb_results, hb_invoke_end)


def _run_local_merge(
    params: CompactPartitionParams,
    uniform_deltas: List[DeltaAnnotated],
    compacted_partition: Partition,
    round_completion_info: RoundCompletionInfo,
    delete_strategy: Optional[DeleteStrategy],
    delete_file_envelopes: Optional[DeleteFileEnvelope],
    mutable_compaction_audit: CompactionSessionAuditInfo,
    previous_compacted_delta_manifest: Optional[Manifest],
    total_input_records_count: np.int64,
) -> tuple[List[MergeResult], np.int64]:
    local_merge_input: MergeInput = generate_local_merge_input(
        params,
        uniform_deltas,
        compacted_partition,
        round_completion_info,
        delete_strategy,
        delete_file_envelopes,
    )
    estimated_da_bytes = (
        mutable_compaction_audit.estimated_in_memory_size_bytes_during_discovery
    )
    estimated_num_records: int = sum(
        [
            entry.meta.record_count
            for delta in uniform_deltas
            for entry in delta.manifest.entries
        ]
    )
    local_merge_options = local_merge_resource_options_provider(
        estimated_da_size=estimated_da_bytes,
        estimated_num_rows=estimated_num_records,
        total_memory_buffer_percentage=params.total_memory_buffer_percentage,
        round_completion_info=round_completion_info,
        compacted_delta_manifest=previous_compacted_delta_manifest,
        ray_custom_resources=params.ray_custom_resources,
        primary_keys=params.primary_keys,
        memory_logs_enabled=params.memory_logs_enabled,
        estimate_resources_params=params.estimate_resources_params,
    )
    local_merge_result = ray.get(
        mg.merge.options(**local_merge_options).remote(local_merge_input)
    )
    total_input_records_count += local_merge_result.input_record_count
    merge_results = [local_merge_result]
    return merge_results, total_input_records_count


def _process_merge_results(
    params: CompactPartitionParams,
    merge_results: List[MergeResult],
    mutable_compaction_audit: CompactionSessionAuditInfo,
) -> tuple[Delta, List[MaterializeResult], dict]:
    mat_results = []
    for merge_result in merge_results:
        mat_results.extend(merge_result.materialize_results)

    mat_results: List[MaterializeResult] = sorted(
        mat_results, key=lambda m: m.task_index
    )
    hb_id_to_entry_indices_range = {}
    file_index = 0
    previous_task_index = -1

    duplicate_hash_bucket_mat_results = 0
    for mat_result in mat_results:
        assert (
            mat_result.pyarrow_write_result.files >= 1
        ), "At least one file must be materialized"
        if mat_result.task_index == previous_task_index:
            duplicate_hash_bucket_mat_results += 1
        else:
            duplicate_hash_bucket_mat_results = 0
        assert duplicate_hash_bucket_mat_results < params.num_rounds, (
            f"Duplicate record count ({duplicate_hash_bucket_mat_results}) is as large "
            f"as or greater than params.num_rounds, which is {params.num_rounds}"
        )
        # ensure start index is the first file index if task index is same
        hb_id_to_entry_indices_range[str(mat_result.task_index)] = (
            hb_id_to_entry_indices_range.get(str(mat_result.task_index), [file_index])[
                0
            ],
            file_index + mat_result.pyarrow_write_result.files,
        )

        file_index += mat_result.pyarrow_write_result.files
        previous_task_index = mat_result.task_index

    _upload_audit_data(
        params,
        mutable_compaction_audit,
    )
    deltas: List[Delta] = [m.delta for m in mat_results]
    # Note: An appropriate last stream position must be set
    # to avoid correctness issue.
    merged_delta: Delta = Delta.merge_deltas(
        deltas,
        stream_position=params.last_stream_position_to_compact,
    )

    return merged_delta, mat_results, hb_id_to_entry_indices_range


def _update_and_upload_compaction_audit(
    params: CompactPartitionParams,
    mutable_compaction_audit: CompactionSessionAuditInfo,
    round_completion_info: Optional[RoundCompletionInfo] = None,
) -> None:

    # After all incremental delta related calculations, we update
    # the input sizes to accommodate the compacted table
    if round_completion_info:
        mutable_compaction_audit.set_input_file_count(
            (mutable_compaction_audit.input_file_count or 0)
            + round_completion_info.compacted_pyarrow_write_result.files
        )
        mutable_compaction_audit.set_input_size_bytes(
            (mutable_compaction_audit.input_size_bytes or 0.0)
            + round_completion_info.compacted_pyarrow_write_result.file_bytes
        )
        mutable_compaction_audit.set_input_records(
            (mutable_compaction_audit.input_records or 0)
            + round_completion_info.compacted_pyarrow_write_result.records
        )

    _upload_audit_data(
        params,
        mutable_compaction_audit,
    )
    return


def _create_round_completion_info(
    params: CompactPartitionParams,
    mutable_compaction_audit: CompactionSessionAuditInfo,
    compacted_partition: Partition,
    audit_url: str,
    hb_id_to_entry_indices_range: dict,
    rci_source_partition_locator: PartitionLocator,
    new_compacted_delta_locator: DeltaLocator,
    pyarrow_write_result: PyArrowWriteResult,
    prev_round_completion_info: Optional[RoundCompletionInfo] = None,
) -> ExecutionCompactionResult:
    input_inflation = None
    input_average_record_size_bytes = None
    # Note: we only consider inflation for incremental delta
    if (
        mutable_compaction_audit.input_size_bytes
        and mutable_compaction_audit.hash_bucket_processed_size_bytes
    ):
        input_inflation = (
            mutable_compaction_audit.hash_bucket_processed_size_bytes
            / mutable_compaction_audit.input_size_bytes
        )

    if (
        mutable_compaction_audit.hash_bucket_processed_size_bytes
        and mutable_compaction_audit.input_records
    ):
        input_average_record_size_bytes = (
            mutable_compaction_audit.hash_bucket_processed_size_bytes
            / mutable_compaction_audit.input_records
        )

    logger.info(
        f"The inflation of input deltas={input_inflation}"
        f" and average record size={input_average_record_size_bytes}"
    )

    mutable_compaction_audit.set_observed_input_inflation(input_inflation)
    mutable_compaction_audit.set_observed_input_average_record_size_bytes(
        input_average_record_size_bytes
    )

    _update_and_upload_compaction_audit(
        params,
        mutable_compaction_audit,
        prev_round_completion_info,
    )

    # Check if this is an in-place compaction before creating RoundCompletionInfo
    logger.info(
        f"Checking if partition {rci_source_partition_locator} is inplace compacted against {params.destination_partition_locator}..."
    )
    is_inplace_compacted: bool = _is_inplace_compacted(
        rci_source_partition_locator, params.destination_partition_locator
    )

    # Determine the prev_source_partition_locator based on compaction type
    if is_inplace_compacted:
        logger.info(
            "In-place compaction detected. Using compacted partition locator as prev_source_partition_locator. "
            + f"Got compacted partition partition_id of {compacted_partition.locator.partition_id} "
            f"and rci source partition_id of {rci_source_partition_locator.partition_id}."
        )
        prev_source_partition_locator = compacted_partition.locator
        # Update rci_source_partition_locator for backward compatibility
        rci_source_partition_locator = compacted_partition.locator
    else:
        prev_source_partition_locator = rci_source_partition_locator

    new_round_completion_info = RoundCompletionInfo.of(
        high_watermark=params.last_stream_position_to_compact,
        compacted_delta_locator=new_compacted_delta_locator,
        compacted_pyarrow_write_result=pyarrow_write_result,
        sort_keys_bit_width=params.bit_width_of_sort_keys,
        manifest_entry_copied_by_reference_ratio=mutable_compaction_audit.untouched_file_ratio,
        compaction_audit_url=audit_url,
        hash_bucket_count=params.hash_bucket_count,
        hb_index_to_entry_range=hb_id_to_entry_indices_range,
        compactor_version=CompactorVersion.V2.value,
        input_inflation=input_inflation,
        input_average_record_size_bytes=input_average_record_size_bytes,
        prev_source_partition_locator=prev_source_partition_locator,
    )

    logger.info(
        f"Partition-{params.source_partition_locator.partition_values},"
        f"compacted at: {params.last_stream_position_to_compact},"
    )

    return ExecutionCompactionResult(
        compacted_partition,
        new_round_completion_info,
        is_inplace_compacted,
    )


def _commit_compaction_result(
    params: CompactPartitionParams,
    execute_compaction_result: ExecutionCompactionResult,
) -> None:
    compaction_session_type: str = (
        "INPLACE" if execute_compaction_result.is_inplace_compacted else "NON-INPLACE"
    )
    logger.info(
        f"Partition-{params.source_partition_locator} -> "
        f"{compaction_session_type} Compaction session data processing completed"
    )
    # TODO(pdames): Uncomment this once we support concurrent writes to the same
    #   partition (via write_to_table). This requires updating the commit_partition
    #   method to support previous partition as input. Right now, a concurrent write
    #   to the same partition will cause the commit_partition method to fail.
    if execute_compaction_result.new_compacted_partition:
        previous_partition: Optional[Partition] = None
        #   if execute_compaction_result.is_inplace_compacted:
        #       previous_partition: Optional[
        #           Partition
        #       ] = params.deltacat_storage.get_partition(
        #           params.source_partition_locator.stream_locator,
        #           params.source_partition_locator.partition_values,
        #           **params.deltacat_storage_kwargs,
        #       )
        #       # NOTE: Retrieving the previous partition again as the partition_id may have changed by the time commit_partition is called.
        logger.info(
            f"Committing compacted partition to: {execute_compaction_result.new_compacted_partition.locator} "
            f"using previous partition: {previous_partition.locator if previous_partition else None}"
        )
        # Set the round completion info on the partition before committing
        execute_compaction_result.new_compacted_partition.compaction_round_completion_info = (
            execute_compaction_result.new_round_completion_info
        )
        committed_partition: Partition = params.deltacat_storage.commit_partition(
            execute_compaction_result.new_compacted_partition,
            previous_partition,
            **params.deltacat_storage_kwargs,
        )
        logger.info(f"Committed compacted partition: {committed_partition}")
    else:
        logger.warning("No new partition was committed during compaction.")

    logger.info(f"Completed compaction session for: {params.source_partition_locator}")


def _upload_audit_data(
    params: CompactPartitionParams,
    audit_info: CompactionSessionAuditInfo,
) -> None:
    """
    Upload audit data to the specified URL using the filesystem from catalog properties.
    """
    audit_url = audit_info.audit_url
    audit_data = json.dumps(audit_info.to_serializable(params.catalog.root))
    if params.catalog and params.catalog.filesystem:
        # Use the filesystem from catalog properties
        filesystem = params.catalog.filesystem
        parsed_url = urlparse(audit_url)
        # For filesystem paths, use the path component
        path = parsed_url.path if parsed_url.scheme else audit_url

        # Ensure parent directories exist
        import os

        parent_dir = os.path.dirname(path)
        if (
            parent_dir
            and not filesystem.get_file_info(parent_dir).type
            == pyarrow.fs.FileType.Directory
        ):
            try:
                filesystem.create_dir(parent_dir, recursive=True)
            except Exception as e:
                logger.warning(f"Failed to create directory {parent_dir}: {e}")

        with filesystem.open_output_stream(path) as output_stream:
            output_stream.write(audit_data.encode("utf-8"))
    else:
        # Fallback: resolve filesystem from the URL
        path, filesystem = resolve_path_and_filesystem(audit_url)

        # Ensure parent directories exist
        import os

        parent_dir = os.path.dirname(path)
        if (
            parent_dir
            and not filesystem.get_file_info(parent_dir).type
            == pyarrow.fs.FileType.Directory
        ):
            try:
                filesystem.create_dir(parent_dir, recursive=True)
            except Exception as e:
                logger.warning(f"Failed to create directory {parent_dir}: {e}")

        with filesystem.open_output_stream(path) as output_stream:
            output_stream.write(audit_data.encode("utf-8"))



from deltacat.experimental.storage.rivulet.fs.file_provider import FileProvider
from deltacat.experimental.storage.rivulet.metastore.sst import SSTableRow
from deltacat.experimental.storage.rivulet.reader.data_reader import FileReader
from typing import Type, Dict

from deltacat.experimental.storage.rivulet.schema.schema import Schema


class FileReaderRegistrar:
    """
    Registrar for readers of rivulet data

    Readers must adhere to the Protocol DataReader

    Packages with extension classes should call into this registrar in __init__.py
    """

    _readers = {}

    @classmethod
    def register_reader(
        cls,
        extension: str,
        reader_class: Type[FileReader],
        allow_overwrite: bool = False,
    ):
        """
        Register a file extension associated with a dataset reader

        Parameters:
        - extension: str, the file extension to register
        - reader_class: Type[DataReader], the reader class to associate with the extension
        - allow_overwrite: bool, if True, allows overwriting an existing reader for the extension
        """
        if extension in cls._readers and not allow_overwrite:
            raise ValueError(
                f"Reader for extension '{extension}' is already registered. "
                f"Set allow_overwrite=True to replace the existing reader."
            )
        normalized_extension = extension.lower()
        cls._readers[normalized_extension] = reader_class

    @classmethod
    def get_reader_class(cls, uri: str) -> Type[FileReader]:
        """
        Gets the reader class given a URI

        :param uri: URI of file to be read. Note that we expect the URI to end in a file extension
        :raises ValueError: if no registered data reader is found for the URI's extension type
        """
        # Find the file extension from the URI
        extension = uri.split(".")[-1].lower()

        # Return the reader class if the extension is registered, otherwise return None
        return cls._readers.get(extension)

    @classmethod
    def construct_reader_instance(
        cls,
        sst_row: SSTableRow,
        file_provider: FileProvider,
        primary_key: str,
        schema: Schema,
        reader_cache: Dict[str, FileReader] = None,
    ) -> FileReader:
        """
        Construct a data reader for an instance of a given uri

        :param uri: URI of file to be read. Note that we expect the URI to end in a file extension
        :param reader_cache: Optional cache of readers keyed by extension
        :raises ValueError: if no registered data reader is found for the URI's extension type
        """
        extension = sst_row.uri.split(".")[-1].lower()

        if reader_cache is not None and extension in reader_cache:
            return reader_cache[extension]

        reader_class = FileReaderRegistrar.get_reader_class(sst_row.uri)
        reader_instance = reader_class(sst_row, file_provider, primary_key, schema)

        if reader_cache:
            reader_cache[extension] = reader_instance

        return reader_instance



import importlib
import logging
import time
from contextlib import nullcontext
from typing import List, Optional, Tuple
from deltacat.compute.compactor_v2.model.hash_bucket_input import HashBucketInput
import numpy as np
import ray
from deltacat import logs
from deltacat.compute.compactor import (
    DeltaAnnotated,
    DeltaFileEnvelope,
)
from deltacat.compute.compactor.model.delta_file_envelope import DeltaFileEnvelopeGroups
from deltacat.compute.compactor_v2.model.hash_bucket_result import HashBucketResult
from deltacat.compute.compactor_v2.utils.delta import read_delta_file_envelopes
from deltacat.compute.compactor_v2.utils.primary_key_index import (
    group_hash_bucket_indices,
    group_by_pk_hash_bucket,
)
from deltacat.storage import metastore
from deltacat.utils.ray_utils.runtime import (
    get_current_ray_task_id,
    get_current_ray_worker_id,
)
from deltacat.utils.common import ReadKwargsProvider
from deltacat.utils.performance import timed_invocation
from deltacat.utils.metrics import emit_timer_metrics, failure_metric, success_metric
from deltacat.utils.resources import (
    get_current_process_peak_memory_usage_in_bytes,
    ProcessUtilizationOverTimeRange,
)
from deltacat.constants import BYTES_PER_GIBIBYTE
from deltacat.compute.compactor_v2.constants import (
    HASH_BUCKET_TIME_IN_SECONDS,
    HASH_BUCKET_FAILURE_COUNT,
    HASH_BUCKET_SUCCESS_COUNT,
)
from deltacat.exceptions import (
    categorize_errors,
)

if importlib.util.find_spec("memray"):
    import memray

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def _group_file_records_by_pk_hash_bucket(
    annotated_delta: DeltaAnnotated,
    num_hash_buckets: int,
    primary_keys: List[str],
    all_column_names: List[str],
    read_kwargs_provider: Optional[ReadKwargsProvider] = None,
    deltacat_storage=metastore,
    deltacat_storage_kwargs: Optional[dict] = None,
) -> Tuple[Optional[DeltaFileEnvelopeGroups], int, int]:
    # read input parquet s3 objects into a list of delta file envelopes
    (
        delta_file_envelopes,
        total_record_count,
        total_size_bytes,
    ) = read_delta_file_envelopes(
        annotated_delta,
        all_column_names,
        read_kwargs_provider,
        deltacat_storage,
        deltacat_storage_kwargs,
    )

    if delta_file_envelopes is None:
        return None, 0, 0

    logger.info(
        f"Read all delta file envelopes: {len(delta_file_envelopes)} "
        f"and total_size_bytes={total_size_bytes} and records={total_record_count}"
    )

    # group the data by primary key hash value
    hb_to_delta_file_envelopes = np.empty([num_hash_buckets], dtype="object")
    for dfe in delta_file_envelopes:
        logger.info("Grouping by pk hash bucket")
        group_start = time.monotonic()
        hash_bucket_to_table = group_by_pk_hash_bucket(
            table=dfe.table, num_buckets=num_hash_buckets, primary_keys=primary_keys
        )
        group_end = time.monotonic()
        logger.info(f"Grouping took: {group_end - group_start}")
        for hb, table in enumerate(hash_bucket_to_table):
            if table:
                if hb_to_delta_file_envelopes[hb] is None:
                    hb_to_delta_file_envelopes[hb] = []
                hb_to_delta_file_envelopes[hb].append(
                    DeltaFileEnvelope.of(
                        stream_position=dfe.stream_position,
                        file_index=dfe.file_index,
                        delta_type=dfe.delta_type,
                        table=table,
                    )
                )
    return hb_to_delta_file_envelopes, total_record_count, total_size_bytes


@success_metric(name=HASH_BUCKET_SUCCESS_COUNT)
@failure_metric(name=HASH_BUCKET_FAILURE_COUNT)
@categorize_errors
def _timed_hash_bucket(input: HashBucketInput):
    task_id = get_current_ray_task_id()
    worker_id = get_current_ray_worker_id()
    with memray.Tracker(
        f"hash_bucket_{worker_id}_{task_id}.bin"
    ) if input.enable_profiler else nullcontext():
        (
            delta_file_envelope_groups,
            total_record_count,
            total_size_bytes,
        ) = _group_file_records_by_pk_hash_bucket(
            annotated_delta=input.annotated_delta,
            num_hash_buckets=input.num_hash_buckets,
            primary_keys=input.primary_keys,
            all_column_names=input.all_column_names,
            read_kwargs_provider=input.read_kwargs_provider,
            deltacat_storage=input.deltacat_storage,
            deltacat_storage_kwargs=input.deltacat_storage_kwargs,
        )
        hash_bucket_group_to_obj_id_tuple = group_hash_bucket_indices(
            hash_bucket_object_groups=delta_file_envelope_groups,
            num_buckets=input.num_hash_buckets,
            num_groups=input.num_hash_groups,
            object_store=input.object_store,
        )

        peak_memory_usage_bytes = get_current_process_peak_memory_usage_in_bytes()
        logger.info(
            f"Peak memory usage in bytes after hash bucketing: {peak_memory_usage_bytes}"
        )
        return HashBucketResult(
            hash_bucket_group_to_obj_id_tuple,
            np.int64(total_size_bytes),
            np.int64(total_record_count),
            np.double(peak_memory_usage_bytes),
            np.double(0.0),
            np.double(time.time()),
        )


@ray.remote
def hash_bucket(input: HashBucketInput) -> HashBucketResult:
    with ProcessUtilizationOverTimeRange() as process_util:
        logger.info(f"Starting hash bucket task {input.hb_task_index}...")

        # Log node peak memory utilization every 10 seconds
        def log_peak_memory():
            logger.debug(
                f"Process peak memory utilization so far: {process_util.max_memory} bytes "
                f"({process_util.max_memory/BYTES_PER_GIBIBYTE} GB)"
            )

        if input.memory_logs_enabled:
            process_util.schedule_callback(log_peak_memory, 10)

        hash_bucket_result, duration = timed_invocation(
            func=_timed_hash_bucket, input=input
        )

        emit_metrics_time = 0.0
        if input.metrics_config:
            emit_result, latency = timed_invocation(
                func=emit_timer_metrics,
                metrics_name=HASH_BUCKET_TIME_IN_SECONDS,
                value=duration,
                metrics_config=input.metrics_config,
            )
            emit_metrics_time = latency

        logger.info(f"Finished hash bucket task {input.hb_task_index}...")
        return HashBucketResult(
            hash_bucket_result[0],
            hash_bucket_result[1],
            hash_bucket_result[2],
            hash_bucket_result[3],
            np.double(emit_metrics_time),
            hash_bucket_result[5],
        )



import itertools

from deltacat.compute.compactor_v2.deletes.model import (
    PrepareDeleteResult,
    DeleteFileEnvelope,
)
from deltacat.storage import (
    DeltaType,
)
from collections import defaultdict
import logging

from typing import Optional, List, Dict, Tuple
from deltacat.types.media import StorageType
from deltacat.compute.compactor.model.compact_partition_params import (
    CompactPartitionParams,
)
from deltacat.compute.compactor_v2.deletes.delete_strategy_equality_delete import (
    EqualityDeleteStrategy,
)
import pyarrow as pa
from deltacat.storage import (
    Delta,
)
from deltacat import logs
from deltacat.utils.metrics import metrics
from deltacat.compute.compactor_v2.constants import PREPARE_DELETES_METRIC_PREFIX


logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def _aggregate_delete_deltas(input_deltas: List[Delta]) -> Dict[int, List[Delta]]:
    """
    Aggregates consecutive DELETE deltas with the same delete parameters into groups.

    Args:
        input_deltas (List[Delta]): A list of Delta objects representing delete operations.
    Returns:
        Dict[int, List[Delta]]: A dictionary where the keys are the stream positions of the
        earliest delta in each group of consecutive DELETE deltas with the same delete parameters,
        and the values are lists containing those deltas.
    """
    start_stream_spos_to_delete_delta_sequence: Dict[int, List[Delta]] = defaultdict(
        list
    )
    delete_deltas_sequence_grouped_by_delete_parameters: List[
        Tuple[bool, List[Delta]]
    ] = [
        (is_delete, list(delete_delta_group))
        for (is_delete, _), delete_delta_group in itertools.groupby(
            input_deltas, lambda d: (d.type is DeltaType.DELETE, d.meta.entry_params)
        )
    ]
    for (
        is_delete_delta_sequence,
        delete_delta_sequence,
    ) in delete_deltas_sequence_grouped_by_delete_parameters:
        if not is_delete_delta_sequence:
            continue
        starting_stream_position_of_delete_sequence: int = delete_delta_sequence[
            0
        ].stream_position
        start_stream_spos_to_delete_delta_sequence[
            starting_stream_position_of_delete_sequence
        ] = delete_delta_sequence
    return start_stream_spos_to_delete_delta_sequence


def _get_delete_file_envelopes(
    params: CompactPartitionParams,
    delete_spos_to_delete_deltas: Dict[int, List],
) -> List[DeleteFileEnvelope]:
    """
    Create a list of DeleteFileEnvelope objects from the given dictionary of delete deltas.
    Args:
        params (CompactPartitionParams): compaction session parameters
        delete_spos_to_delete_deltas (Dict[int, List]): A dictionary where the keys are the stream positions of the
            earliest delta in each group of consecutive DELETE deltas with the same delete parameters,
            and the values are lists containing those deltas.
    Returns:
        List[DeleteFileEnvelope]: A list of DeleteFileEnvelope objects.
    """
    delete_file_envelopes = []
    for (
        start_stream_position,
        delete_delta_sequence,
    ) in delete_spos_to_delete_deltas.items():
        consecutive_delete_tables: List[pa.Table] = []
        for delete_delta in delete_delta_sequence:
            assert (
                delete_delta.meta.entry_params is not None
            ), "Delete type deltas are required to have delete parameters defined"
            delete_columns: Optional[
                List[str]
            ] = delete_delta.meta.entry_params.equality_field_locators
            assert len(delete_columns) > 0, "At least 1 delete column is required"
            # delete columns should exist in underlying table
            delete_dataset = params.deltacat_storage.download_delta(
                delete_delta,
                file_reader_kwargs_provider=params.read_kwargs_provider,
                columns=delete_columns,
                storage_type=StorageType.LOCAL,
                max_parallelism=1,
                **params.deltacat_storage_kwargs,
            )
            consecutive_delete_tables.extend(delete_dataset)
        delete_table: pa.Table = pa.concat_tables(consecutive_delete_tables)
        delete_file_envelopes.append(
            DeleteFileEnvelope.of(
                start_stream_position,
                delta_type=DeltaType.DELETE,
                table=delete_table,
                delete_columns=delete_columns,
            )
        )
    return delete_file_envelopes


@metrics(prefix=PREPARE_DELETES_METRIC_PREFIX)
def prepare_deletes(
    params: CompactPartitionParams,
    input_deltas: List[Delta],
    *args,
    **kwargs,
) -> PrepareDeleteResult:
    """
    Prepares delete operations for a compaction process.

    This function processes all the deltas and consolidates consecutive DELETE deltas.
    It creates a list of these delete file envelopes.
    Additionally, non-DELETE deltas are accumulated in a separate list.

    Args:
        params (CompactPartitionParams): Parameters for the compaction process.
        input_deltas (List[Delta]): A list of Delta objects representing delete operations.

    Returns:
        PrepareDeleteResult:
            - A list of Deltas excluding all DELETE operations.
            - A list of DeleteFileEnvelope objects representing the consolidated delete operations.
            - An instance of the EqualityDeleteStrategy class.

    Raises:
        AssertionError: If the input_deltas list is not sorted in non-decreasing order by stream_position.
        AssertionError: If the number of delete file envelopes does not match the number of DELETE-type Delta sequences.
    """
    if not input_deltas:
        return PrepareDeleteResult(input_deltas, [], None)
    assert all(
        input_deltas[i].stream_position <= input_deltas[i + 1].stream_position
        for i in range(len(input_deltas) - 1)
    ), "Uniform deltas must be in non-decreasing order by stream position"
    start_stream_spos_to_delete_delta_sequence: Dict[
        int, List[Delta]
    ] = _aggregate_delete_deltas(input_deltas)
    delete_file_envelopes: List[DeleteFileEnvelope] = _get_delete_file_envelopes(
        params, start_stream_spos_to_delete_delta_sequence
    )
    assert len(start_stream_spos_to_delete_delta_sequence) == len(
        delete_file_envelopes
    ), "The number of delete file envelopes should match the number of DELETE-type Delta sequences"
    return PrepareDeleteResult(
        [delta for delta in input_deltas if delta.type is not DeltaType.DELETE],
        delete_file_envelopes,
        EqualityDeleteStrategy(),
    )



import ray
import deltacat as dc
import daft


def print_package_version_info():
    print(f"DeltaCAT Version: {dc.__version__}")
    print(f"Ray Version: {ray.__version__}")
    print(f"Daft Version: {daft.__version__}")


@ray.remote
def hello_worker():
    print("Hello, Worker!")
    df = daft.from_pydict({"hello": ["delta", "cat"]})
    dc.write(df, "hello_world")
    print_package_version_info()


def run():
    print("Hello, Driver!")
    print_package_version_info()
    ray.get(hello_worker.remote())
    df = dc.read("hello_world")
    print("=== Table Written by Ray Worker ===")
    print(df)


if __name__ == "__main__":
    # initialize deltacat
    # Catalog files will be stored in .deltacat/ in the current working directory.
    dc.init_local()

    # run the example
    run()



"""
Exception classes for main storage testing that mirror the local storage exceptions.
These are used to test the main metastore error categorization functionality.
"""


class InvalidNamespaceError(Exception):
    """Exception raised when an invalid namespace is provided to main storage."""

    error_name = "InvalidNamespaceError"


class MainStorageValidationError(Exception):
    """Exception raised when main storage validation fails."""

    error_name = "MainStorageValidationError"


class MainStorageError(Exception):
    """General exception for main storage operations."""

    error_name = "MainStorageError"



import posixpath
import time
from typing import List, Generator

from deltacat.storage.model.partition import PartitionLocator
from deltacat.experimental.storage.rivulet.fs.file_store import FileStore
from deltacat.experimental.storage.rivulet.fs.input_file import InputFile
from deltacat.experimental.storage.rivulet.fs.output_file import OutputFile
from deltacat.utils.metafile_locator import _find_partition_path


class FileProvider:
    """
    Manages the generation of URIs for data and metadata files and facilitates the creation of files at those URIs.
    All files are generated relative to the root of the storage location.

    This class is inspired by the Iceberg `LocationProvider` and provides methods
    to generate paths for various types of files (e.g., data files, SSTs, and manifests)
    while maintaining a clear structure within the dataset.

    TODO (deltacat): FileProvider will be replaced/refactored once we are able to integrate with Deltacat.
    TODO: Incorporate additional file naming conventions, such as including
          partitionId, taskId, and operationId, to improve traceability and
          idempotency.
    """

    uri: str

    def __init__(self, uri: str, locator: PartitionLocator, file_store: FileStore):
        """
        Initializes the file provider.

        param: uri: Base URI of the dataset.
        param: file_store: FileStore instance for creating and reading files.
        """
        self.uri = uri
        self._locator = locator
        self._file_store = file_store

    def provide_data_file(self, extension: str) -> OutputFile:
        """
        Creates a new data file.

        TODO: Ensure storage interface can provide data files.

        param: extension: File extension (e.g., "parquet").
        returns: OutputFile instance pointing to the created data file.
        """
        partition_path = _find_partition_path(self.uri, self._locator)
        uri = posixpath.join(
            partition_path, "data", f"{int(time.time_ns())}.{extension}"
        )
        return self._file_store.create_output_file(uri)

    def provide_l0_sst_file(self) -> OutputFile:
        """
        Creates a new L0 SST file.

        TODO: Ensure storage interface can provide sst files.

        returns: OutputFile instance pointing to the created SST file.
        """
        partition_path = _find_partition_path(self.uri, self._locator)
        uri = posixpath.join(
            partition_path, "metadata", "ssts", "0", f"{int(time.time_ns())}.json"
        )
        return self._file_store.create_output_file(uri)

    def provide_input_file(self, uri: str) -> InputFile:
        """
        Reads an existing file.

        param sst_uri: URI of the file to read.
        returns: InputFile instance for the specified URI.
        """
        return self._file_store.create_input_file(uri)

    def provide_manifest_file(self) -> OutputFile:
        """
        Creates a new manifest file.

        returns: OutputFile instance pointing to the created manifest file.
        """
        uri = f"{self.uri}/metadata/manifests/{int(time.time_ns())}.json"
        return self._file_store.create_output_file(uri)

    def get_sst_scan_directories(self) -> List[str]:
        """
        Retrieves SST scan directories.

        returns: List of directories containing SSTs.
        """
        partition_path = _find_partition_path(self.uri, self._locator)
        return [f"{partition_path}/metadata/ssts/0/"]

    def generate_sst_uris(self) -> Generator[InputFile, None, None]:
        """
        Generates all SST URIs.

        returns: Generator of InputFile instances for SSTs.
        """
        sst_directories = self.get_sst_scan_directories()
        for directory in sst_directories:
            for file in self._file_store.list_files(directory):
                yield file



from abc import ABC, abstractmethod
from contextlib import contextmanager
import posixpath
from typing import Protocol

from pyarrow.fs import FileSystem, FileType

from deltacat.experimental.storage.rivulet.fs.input_file import FSInputFile, InputFile


class OutputStream(Protocol):  # pragma: no cover
    """A protocol with a subset of IOBase for file-like output objects"""

    @abstractmethod
    def write(self, b: bytes) -> int:
        ...

    @abstractmethod
    def close(self) -> None:
        ...

    @abstractmethod
    def __enter__(self) -> "OutputStream":
        ...

    @abstractmethod
    def __exit__(self, exc_type, exc_value, traceback) -> None:
        ...


class OutputFile(ABC):
    """Abstraction for interacting with output files"""

    def __init__(self, location: str):
        self._location = location

    @property
    def location(self) -> str:
        return self._location

    @abstractmethod
    def exists(self) -> bool:
        """Return whether the location exists.

        Raises:
            PermissionError: If this has insufficient permissions to access the file at location.
        """

    @abstractmethod
    def to_input_file(self) -> InputFile:
        """Return an InputFile for this output file's location"""

    @abstractmethod
    def create(self) -> OutputStream:
        """Return a file-like object for output

        TODO: overwrite protection (FileExistsError?)
        Raises:
            PermissionError: If this has insufficient permissions to access the file at location.
        """


class FSOutputFile(OutputFile):
    def __init__(self, location: str, fs: FileSystem):
        self._location = location
        self.fs = fs

    def exists(self) -> bool:
        file_info = self.fs.get_file_info(self._location)
        return file_info.type != FileType.NotFound

    def to_input_file(self) -> "FSInputFile":
        return FSInputFile(self._location, self.fs)

    @contextmanager
    def create(self):
        """Create and open the file for writing."""
        try:
            parent_dir = posixpath.dirname(self._location)
            if parent_dir:  # Check if there's a parent directory to create
                self.fs.create_dir(parent_dir, recursive=True)

            with self.fs.open_output_stream(self._location) as output_stream:
                yield output_stream
        except Exception as e:
            raise IOError(f"Failed to create or write to file '{self._location}': {e}")
